"""Perfect nested sampling generators."""
import numpy as np
from anesthetic.examples.utils import random_ellipsoid
from anesthetic import NestedSamples
from anesthetic.samples import merge_nested_samples


def gaussian(nlive, ndims, sigma=0.1, R=1, logLmin=-1e-2):
    """Perfect nested sampling run for a spherical Gaussian & prior.

    Up to normalisation this is identical to the example in John Skilling's
    original paper https://doi.org/10.1214/06-BA127 . Both spherical Gaussian
    width sigma and spherical uniform prior width R are centered on zero

    Parameters
    ----------
    nlive : int
        number of live points

    ndims : int
        dimensionality of gaussian

    sigma : float
        width of gaussian likelihood

    R : float
        radius of gaussian prior

    logLmin : float
        loglikelihood at which to terminate

    Returns
    -------
    samples : :class:`anesthetic.samples.NestedSamples`
        Nested sampling run
    """

    def logLike(x):
        return -(x**2).sum(axis=-1)/2/sigma**2

    def random_sphere(n):
        return random_ellipsoid(np.zeros(ndims), np.eye(ndims), n)

    samples = []
    r = R
    logL_birth = np.ones(nlive) * -np.inf
    logL = logL_birth.copy()
    while logL.min() < logLmin:
        points = r * random_sphere(nlive)
        logL = logLike(points)
        samples.append(NestedSamples(points, logL=logL, logL_birth=logL_birth))
        logL_birth = logL.copy()
        r = (points**2).sum(axis=-1, keepdims=True)**0.5

    samples = merge_nested_samples(samples)
    logLend = samples.loc[samples.nlive >= nlive].logL.max()
    return samples.loc[samples.logL_birth < logLend].recompute()


def correlated_gaussian(nlive, mean, cov, bounds=None):
    """Perfect nested sampling run for a correlated gaussian likelihood.

    This produces a perfect nested sampling run with a uniform prior over the
    unit hypercube, with a likelihood gaussian in the parameters normalised so
    that the evidence is unity. The algorithm proceeds by simultaneously
    rejection sampling from the prior and sampling exactly and uniformly from
    the known ellipsoidal contours.

    This can produce perfect runs in up to around d~15 dimensions. Beyond
    this rejection sampling from a truncated gaussian in the early stage
    becomes too inefficient.

    Parameters
    ----------
    nlive : int
        minimum number of live points across the run

    mean : 1d array-like, shape (ndims,)
        mean of gaussian in parameters. Length of array defines dimensionality
        of run.

    cov : 2d array-like, shape (ndims, ndims)
        covariance of gaussian in parameters

    bounds : 2d array-like, shape (ndims, 2)
        bounds of a gaussian, default ``[[0, 1]]*ndims``

    Returns
    -------
    samples : :class:`anesthetic.samples.NestedSamples`
        Nested sampling run
    """
    mean = np.array(mean, dtype=float)
    cov = np.array(cov, dtype=float)
    invcov = np.linalg.inv(cov)

    def logLike(x):
        return -0.5 * ((x-mean) @ invcov * (x-mean)).sum(axis=-1)

    ndims = len(mean)

    if bounds is None:
        bounds = [[0, 1]]*ndims

    bounds = np.array(bounds, dtype=float)

    logLmax = logLike(mean)

    points = np.random.uniform(*bounds.T, (2*nlive, ndims))
    samples = NestedSamples(points, logL=logLike(points), logL_birth=-np.inf)

    while (1/samples.nlive.iloc[:-nlive]).sum() < samples.D_KL()*2:
        logLs = samples.logL.iloc[-nlive]

        sig = np.diag(cov*2*(logLmax - logLs))**0.5
        bounds[:, 0] = np.max([bounds[:, 0], mean - sig], axis=0)
        bounds[:, 1] = np.min([bounds[:, 1], mean + sig], axis=0)

        # Cube round
        points = np.random.uniform(*bounds.T, (nlive, ndims))
        logL = logLike(points)
        i = logL > logLs
        samps_1 = NestedSamples(points[i], logL=logL[i], logL_birth=logLs)

        # Ellipsoidal round
        points = random_ellipsoid(mean, cov*2*(logLmax - logLs), nlive)
        logL = logLike(points)
        i = ((points > bounds.T[0]) & (points < bounds.T[1])).all(axis=1)
        samps_2 = NestedSamples(points[i], logL=logL[i], logL_birth=logLs)
        print(len(samps_1), len(samps_2))

        samples = merge_nested_samples([samples, samps_1, samps_2])

    return samples


def wedding_cake(nlive, ndims, sigma=0.01, alpha=0.5):
    """Perfect nested sampling run for a wedding cake likelihood.

    This is a likelihood with nested hypercuboidal plateau regions of constant
    likelihood centered on 0.5, with geometrically decreasing volume by a
    factor of alpha. The value of the likelihood in these plateau regions has a
    gaussian profile with width sigma.

    ::

        logL = - alpha^(2*floor(D*log_alpha(2|x-0.5|_infinity))/D) / (8sigma^2)

    Parameters
    ----------
    nlive : int
        number of live points

    ndims : int
        dimensionality of the likelihood

    sigma : float
        width of gaussian profile

    alpha : float
        volume compression between plateau regions
    """

    def i(x):
        """Plateau number of a parameter point."""
        r = np.max(abs(x-0.5), axis=-1)
        return np.floor(ndims*np.log(2*r)/np.log(alpha))

    def logL(x):
        """Gaussian log-likelihood."""
        ri = alpha**(i(x)/ndims)/2
        return - ri**2/2/sigma**2

    points = np.zeros((0, ndims))
    death_likes = np.zeros(0)
    birth_likes = np.zeros(0)

    live_points = np.random.rand(nlive, ndims)
    live_likes = logL(live_points)
    live_birth_likes = np.ones(nlive)*-np.inf

    while True:
        logL_ = live_likes.min()
        j = live_likes == logL_

        death_likes = np.concatenate([death_likes, live_likes[j]])
        birth_likes = np.concatenate([birth_likes, live_birth_likes[j]])
        points = np.concatenate([points, live_points[j]])

        i_ = i(live_points[j][0])+1
        r_ = alpha**(i_/ndims)/2
        x_ = np.random.uniform(0.5-r_, 0.5+r_, size=(j.sum(), ndims))
        live_birth_likes[j] = logL_
        live_points[j] = x_
        live_likes[j] = logL(x_)

        samps = NestedSamples(points, logL=death_likes, logL_birth=birth_likes)
        weights = samps.get_weights()
        if weights[-nlive:].sum() < 0.001 * weights.sum():
            break

    death_likes = np.concatenate([death_likes, live_likes])
    birth_likes = np.concatenate([birth_likes, live_birth_likes])
    points = np.concatenate([points, live_points])

    return NestedSamples(points, logL=death_likes, logL_birth=birth_likes)


def planck_gaussian(nlive=500):
    """Gaussian likelihood with Planck-like posterior.

    This is a gaussian likelihood with the same mean, parameter covariance and
    average loglikelihood as the Planck 2018 legacy chains
    ``base/plikHM_TTTEEE_lowl_lowE_lensing``

    Parameters
    ----------
    nlive : int, optional
        number of live points

    Returns
    -------
    samples : :class:`anesthetic.samples.NestedSamples`
        Nested sampling run
    """
    columns = ['omegabh2', 'omegach2', 'theta', 'tau', 'logA', 'ns']
    labels = [r'$\Omega_b h^2$', r'$\Omega_c h^2$', r'$100\theta_{MC}$',
              r'$\tau$', r'${\rm{ln}}(10^{10} A_s)$', r'$n_s$']

    cov = np.array([
        [2.12e-08, -9.03e-08, 1.76e-08, 2.96e-07, 4.97e-07, 2.38e-07],
        [-9.03e-08, 1.39e-06, -1.26e-07, -3.41e-06, -4.15e-06, -3.28e-06],
        [1.76e-08, -1.26e-07, 9.71e-08, 4.30e-07, 7.41e-07, 4.13e-07],
        [2.96e-07, -3.41e-06, 4.30e-07, 5.33e-05, 9.53e-05, 1.05e-05],
        [4.97e-07, -4.15e-06, 7.41e-07, 9.53e-05, 2.00e-04, 1.35e-05],
        [2.38e-07, -3.28e-06, 4.13e-07, 1.05e-05, 1.35e-05, 1.73e-05]])

    mean = np.array([0.02237, 0.1200, 1.04092, 0.0544, 3.044, 0.9649])

    bounds = np.array([
        [5.00e-03, 1.00e-01],
        [1.00e-03, 9.90e-01],
        [5.00e-01, 1.00e+01],
        [1.00e-02, 8.00e-01],
        [1.61e+00, 3.91e+00],
        [8.00e-01, 1.20e+00]])

    logL_mean = -1400.35

    samples = correlated_gaussian(nlive, mean, cov, bounds)

    data = samples.iloc[:, :len(columns)].to_numpy()
    logL = samples['logL'].to_numpy()
    logL_birth = samples['logL_birth'].to_numpy()
    logL_birth += logL_mean - samples.logL.mean()
    logL += logL_mean - samples.logL.mean()
    samples = NestedSamples(data=data, columns=columns, labels=labels,
                            logL=logL, logL_birth=logL_birth)
    return samples



"""Main plotting tools."""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import (GridSpec as GS,
                                 GridSpecFromSubplotSpec as sGS)
from anesthetic.gui.widgets import (Widget, Slider, Button,
                                    RadioButtons, TrianglePlot, CheckButtons)


class Higson(Widget):
    """Higson plot as shown in https://arxiv.org/abs/1703.09701 .

    Attributes
    ----------
        curve : :class:`matplotlib.lines.Line2D`
            points currently plotted as a curve.

        point : :class:`matplotlib.lines.Line2D`
            large indicator point currently plotted on the curve.

    """

    def __init__(self, fig, gridspec):
        super().__init__(fig, gridspec)
        self.ax.set_yticks([])
        self.ax.set_ylim(-0.1, 1.1)
        self.ax.set_ylabel(r'$LX$')
        self.ax.set_xlabel(r'$\ln X$')

        self.curve, = self.ax.plot([None], [None], 'k-')
        self.point, = self.ax.plot([None], [None], 'ko')

    def update(self, logX, LX, i):
        """Update the line and the point in the higson plot.

        Parameters
        ----------
            logX : array-like
                log-volume compression values to plot

            LX : array-like
                Likelihood * volume compression

            i : int
                Current location of higson point

        """
        self.point.set_xdata(logX[[i]])
        self.point.set_ydata(LX[[i]])
        self.curve.set_xdata(logX)
        self.curve.set_ydata(LX)

    def reset_range(self):
        """Reset the ranges of the higson plot."""
        xdata = self.curve.get_xdata()
        xdata = xdata[np.isfinite(xdata)]
        self.ax.set_xlim(xdata.max(), xdata.min())


class Evolution(Slider):
    """Slider controlling the evolution stage of the live points."""

    def __init__(self, fig, gridspec, action, valmax):
        super().__init__(fig, gridspec, action, '', 0, valmax, 0, 'horizontal')
        self.slider.valtext.set_horizontalalignment('right')
        self.slider.valtext.set_position((0.98, 0.5))

    def __call__(self):
        """Return the current iteration as an integer."""
        mx = self.slider.valmax-1
        val = int(super().__call__())
        return min(val, mx)

    def set_text(self, logL, n):
        """Set the text at end of slider.

        Parameters
        ----------
            logL : float
                Current loglikelihood of evolution stage

            n : int
                Current number of live points of evolution stage

        """
        text = r'$\ln L$: %.6g, $n_\mathrm{live}$: %i' % (logL, n)
        return super().set_text(text)


class Temperature(Slider):
    """Logarithmic slider controlling temperature of the posterior points."""

    def __init__(self, fig, gridspec, action):
        super().__init__(fig, gridspec, action, r'$kT$', -1, 5, 0, 'vertical')

    def __call__(self):
        """Return the current temperature."""
        return 10**super().__call__()

    def set_text(self, kT):
        """Set the text at end of slider.

        Parameters
        ----------
            kT : float
                Current temperature of posterior points stage

        """
        text = r'%.2g' % kT
        return super().set_text(text)


class RunPlotter(object):
    """Construct a control panel of information on a nested sampling run.

    Parameters
    ----------
    samples : :class:`anesthetic.samples.NestedSamples`
        The root string for the chains files to be used, or a set of nested
        samples.

    Attributes
    ----------
    samples : :class:`anesthetic.samples.NestedSamples`
        Object for extracting nested sampling data from chains files.

    fig : :class:`matplotlib.figure.Figure`
        Reference to the underlying figure

    triangle : :class:`anesthetic.gui.widgets.TrianglePlot`
        Corner plot of live or posterior samples.

    temperature : :class:`anesthetic.gui.plot.Temperature`
        Slider selecting the posterior temperature.

    evolution : :class:`anesthetic.gui.plot.Evolution`
        Slider selecting the live iteration.

    higson : :class:`anesthetic.gui.plot.Higson`
        Higson plot of posterior weights.

    reset : :class:`anesthetic.gui.widgets.Button`
        Button that resets the parameter ranges.

    reload : :class:`anesthetic.gui.widgets.Button`
        Button that reloads the files.

    type : :class:`anesthetic.gui.widgets.RadioButtons`
        Radio buttons that selects whether to plot live or posteriors.

    param_choice : :class:`anesthetic.gui.widgets.CheckButtons`
        Checkbox that selects which parameters to plot.

    """

    def __init__(self, samples, params=None):
        self.samples = samples

        if params:
            self.params = np.array(params)
        else:
            self.params = np.array(self.samples.drop_labels().columns[:10])

        self.fig = plt.figure()
        self._set_up()
        self.redraw(None)

    def _set_up(self):
        """Draw the control panel.

        We implement the control panel using sequential recursive gridspecs.

        +----------------------------------------------+------+
        |                                              |      |
        |                  gs0[0]                      |gs0[1]|
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        +-------------------------+-----------------+--+------+
        |                         |    gs11[0]      |         |
        |         gs10[0]         +-----------------+  gs1[2] |
        +-------------------------+    gs11[1]      |         |
        |                         +-----------------+         |
        |         gs10[1]         |    gs11[2]      |         |
        +-------------------------+-----------------+---------+

        These variable names are included in the __init__ function, and are
        named with an intuitive Huffman coding.

        """
        gs = GS(2, 1, height_ratios=[3, 1])
        gs0 = sGS(1, 2, width_ratios=[19, 1], subplot_spec=gs[0])
        gs1 = sGS(1, 3, width_ratios=[4, 1, 1], subplot_spec=gs[1])
        gs10 = sGS(2, 1, height_ratios=[1, 4], subplot_spec=gs1[0])
        gs11 = sGS(3, 1, height_ratios=[1, 1, 2], subplot_spec=gs1[1])

        self.triangle = TrianglePlot(self.fig, gs0[0])
        self.temperature = Temperature(self.fig, gs0[1], self.update)
        self.evolution = Evolution(self.fig, gs10[0],
                                   self.update, len(self.samples))
        self.higson = Higson(self.fig, gs10[1])
        self.reset = Button(self.fig, gs11[0],
                            self.reset_range, 'Reset Range')
        self.reload = Button(self.fig, gs11[1],
                             self.reload_file, 'Reload File')
        self.type = RadioButtons(self.fig, gs11[2],
                                 ('live', 'posterior'), self.update)
        self.param_choice = CheckButtons(self.fig, gs1[2],
                                         self.params, self.redraw)

    def redraw(self, _):
        """Redraw the triangle plot upon parameter updating."""
        self.triangle.draw(self.param_choice(),
                           self.samples.get_labels_map())
        self.update(None)
        self.reset_range(None)
        self.fig.tight_layout()
        self.fig.canvas.draw()

    def points(self, label):
        """Get sample coordinates from nested sampling samples.

        Parameters
        ----------
        label : str
            label indicating the coordinate to extract.

        Returns
        -------
        array-like:
            sample 'label'-coordinates.

        """
        if self.type() == 'posterior':
            kT = self.temperature()
            return self.samples.posterior_points(1/kT)[label]
        else:
            i = self.evolution()
            logL = self.samples.logL.iloc[i]
            return self.samples.live_points(logL)[label]

    def update(self, _):
        """Update all the plots upon slider changes."""
        logX = np.log(self.samples.nlive / (self.samples.nlive+1)).cumsum()
        kT = self.temperature()
        LX = self.samples.logL/kT + logX
        LX = np.exp(LX-LX.max())
        i = self.evolution()
        logL = self.samples.logL.iloc[i]
        n = self.samples.nlive.iloc[i]

        self.triangle.update(self.points)

        self.evolution.set_text(logL, n)
        self.temperature.set_text(kT)

        self.higson.update(logX, LX, i)
        self.fig.canvas.draw()

    def reload_file(self, _):
        """Reload the data from file."""
        from anesthetic import read_chains
        self.samples = read_chains(self.samples.root)
        self.evolution.reset_range(valmax=len(self.samples))
        self.update(None)

    def reset_range(self, _):
        """Reset the parameter ranges."""
        self.triangle.reset_range()
        self.higson.reset_range()
        self.fig.canvas.draw()



"""Read MCMCSamples from GetDist chains."""
import os
import re
import numpy as np
from anesthetic.samples import MCMCSamples
from pandas import concat


def read_paramnames(root):
    r"""Read ``<root>.paramnames`` in GetDist format.

    This file should contain one or two columns. The first column indicates
    a reference name for the sample, used as labels in the pandas array.
    The second optional column should include the equivalent axis label,
    possibly in tex, with the understanding that it will be surrounded by
    dollar signs, for example

    ``<root>.paramnames``:
    ::

        a1     a_1
        a2     a_2
        omega  \omega

    """
    try:
        paramnames_file = root + '.paramnames'
        with open(paramnames_file, 'r') as f:
            paramnames = []
            labels = {}
            for line in f:
                line = line.strip().split()
                paramname = line[0].replace('*', '')
                paramnames.append(paramname)
                if len(line) > 1:
                    labels[paramname] = '$' + ' '.join(line[1:]) + '$'
            return paramnames, labels
    except IOError:
        return None, {}


def read_getdist(root, *args, **kwargs):
    """Read <root>_1.txt in GetDist format.

    Returns
    -------
    :class:`anesthetic.samples.MCMCSamples`

    """
    dirname, basename = os.path.split(root)

    files = os.listdir(os.path.dirname(root))
    regex = re.escape(basename) + r'((_|.)([0-9]+)|)\.txt'
    matches = [re.match(regex, f) for f in files]
    chains_files = [(m.group(3), os.path.join(dirname, m.group(0)))
                    for m in matches if m]
    if not chains_files:
        raise FileNotFoundError(dirname + '/' + regex + " not found.")

    columns, labels = read_paramnames(root)
    columns = kwargs.pop('columns', columns)
    labels = kwargs.pop('labels', labels)
    kwargs['label'] = kwargs.get('label', os.path.basename(root))

    samples = []
    for i, chains_file in chains_files:
        data = np.loadtxt(chains_file)
        weights, minuslogL, data = np.split(data, [1, 2], axis=1)
        mcmc = MCMCSamples(data=data, columns=columns,
                           weights=weights.flatten(), logL=-minuslogL,
                           labels=labels, *args, **kwargs)
        mcmc['chain'] = int(i) if i else np.nan
        samples.append(mcmc)

    samples = concat(samples)
    samples.index.names = ['index', 'weights']
    samples.sort_values(by=['chain', 'index'], inplace=True)
    samples.reset_index(inplace=True, drop=True)
    samples.root = root
    samples.label = kwargs['label']

    all_same_chain = np.all(samples.chain == samples.chain.iloc[0])
    if all_same_chain or samples.chain.isna().all():
        samples.drop(columns='chain', inplace=True, level=0)
    elif samples.islabelled():
        samples.set_label('chain', r'$n_\mathrm{chain}$')

    return samples



"""Main classes for the anesthetic module.

- :class:`anesthetic.samples.Samples`
- :class:`anesthetic.samples.MCMCSamples`
- :class:`anesthetic.samples.NestedSamples`
"""
import numpy as np
import scipy
import pandas
import copy
import warnings
from pandas import MultiIndex, Series
from collections.abc import Sequence
from anesthetic.utils import (compute_nlive, compute_insertion_indexes,
                              is_int, logsumexp)
from anesthetic.gui.plot import RunPlotter
from anesthetic.weighted_pandas import WeightedDataFrame, WeightedSeries
from anesthetic.labelled_pandas import LabelledDataFrame, LabelledSeries
from pandas.core.accessor import CachedAccessor
from anesthetic.plot import (make_1d_axes, make_2d_axes,
                             AxesSeries, AxesDataFrame)
import anesthetic.weighted_pandas
from anesthetic.plotting import PlotAccessor
anesthetic.weighted_pandas._WeightedObject.plot =\
    CachedAccessor("plot", PlotAccessor)


class WeightedLabelledDataFrame(WeightedDataFrame, LabelledDataFrame):
    """:class:`pandas.DataFrame` with weights and labels."""

    _metadata = WeightedDataFrame._metadata + LabelledDataFrame._metadata

    def __init__(self, *args, **kwargs):
        labels = kwargs.pop('labels', None)
        if not hasattr(self, '_labels'):
            self._labels = ('weights', 'labels')
        super().__init__(*args, **kwargs)
        if labels is not None:
            if isinstance(labels, dict):
                labels = [labels.get(p, '') for p in self]
            self.set_labels(labels, inplace=True)

    def islabelled(self, axis=1):
        """Search for existence of labels."""
        return super().islabelled(axis=axis)

    def get_labels(self, axis=1):
        """Retrieve labels from an axis."""
        return super().get_labels(axis=axis)

    def get_labels_map(self, axis=1):
        """Retrieve mapping from paramnames to labels from an axis."""
        return super().get_labels_map(axis=axis)

    def get_label(self, param, axis=1):
        """Retrieve mapping from paramnames to labels from an axis."""
        return super().get_label(param, axis=axis)

    def set_label(self, param, value, axis=1):
        """Set a specific label to a specific value on an axis."""
        return super().set_label(param, value, axis=axis, inplace=True)

    def drop_labels(self, axis=1):
        """Drop the labels from an axis if present."""
        return super().drop_labels(axis)

    def set_labels(self, labels, axis=1, inplace=False, level=None):
        """Set labels along an axis."""
        return super().set_labels(labels, axis=axis,
                                  inplace=inplace, level=level)

    @property
    def _constructor(self):
        return WeightedLabelledDataFrame

    @property
    def _constructor_sliced(self):
        return WeightedLabelledSeries


class WeightedLabelledSeries(WeightedSeries, LabelledSeries):
    """Series with weights and labels."""

    _metadata = WeightedSeries._metadata + LabelledSeries._metadata

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_labels'):
            self._labels = ('weights', 'labels')
        super().__init__(*args, **kwargs)

    def set_label(self, param, value, axis=0):
        """Set a specific label to a specific value."""
        return super().set_label(param, value, axis=axis, inplace=True)

    @property
    def _constructor(self):
        return WeightedLabelledSeries

    @property
    def _constructor_expanddim(self):
        return WeightedLabelledDataFrame


class Samples(WeightedLabelledDataFrame):
    """Storage and plotting tools for general samples.

    Extends the :class:`pandas.DataFrame` by providing plotting methods and
    standardising sample storage.

    Example plotting commands include
        - ``samples.plot_1d(['paramA', 'paramB'])``
        - ``samples.plot_2d(['paramA', 'paramB'])``
        - ``samples.plot_2d([['paramA', 'paramB'], ['paramC', 'paramD']])``

    Parameters
    ----------
    data : np.array
        Coordinates of samples. shape = (nsamples, ndims).

    columns : list(str)
        reference names of parameters

    weights : np.array
        weights of samples.

    logL : np.array
        loglikelihoods of samples.

    labels : dict or array-like
        mapping from columns to plotting labels

    label : str
        Legend label

    logzero : float, default=-1e30
        The threshold for `log(0)` values assigned to rejected sample points.
        Anything equal or below this value is set to `-np.inf`.

    """

    _metadata = WeightedLabelledDataFrame._metadata + ['label']

    def __init__(self, *args, **kwargs):
        logzero = kwargs.pop('logzero', -1e30)
        logL = kwargs.pop('logL', None)
        if logL is not None:
            logL = np.array(logL)
            logL = np.where(logL <= logzero, -np.inf, logL)
        self.label = kwargs.pop('label', None)

        super().__init__(*args, **kwargs)

        if logL is not None:
            self['logL'] = logL
            if self.islabelled(axis=1):
                self.set_label('logL', r'$\ln\mathcal{L}$')

    @property
    def _constructor(self):
        return Samples

    plot = CachedAccessor("plot", PlotAccessor)

    def plot_1d(self, axes, *args, **kwargs):
        """Create an array of 1D plots.

        Parameters
        ----------
        axes : plotting axes
            Can be:

            * list(str) or str
            * :class:`pandas.Series` of :class:`matplotlib.axes.Axes`

            If a :class:`pandas.Series` is provided as an existing set of axes,
            then this is used for creating the plot. Otherwise, a new set of
            axes are created using the list or lists of strings.

        kind : str, default='kde_1d'
            What kind of plots to produce. Alongside the usual pandas options
            {'hist', 'box', 'kde', 'density'}, anesthetic also provides

            * 'hist_1d': :func:`anesthetic.plot.hist_plot_1d`
            * 'kde_1d': :func:`anesthetic.plot.kde_plot_1d`
            * 'fastkde_1d': :func:`anesthetic.plot.fastkde_plot_1d`

            Warning -- while the other pandas plotting options
            {'line', 'bar', 'barh', 'area', 'pie'} are also accessible, these
            can be hard to interpret/expensive for :class:`Samples`,
            :class:`MCMCSamples`, or :class:`NestedSamples`.

        Returns
        -------
        axes : :class:`pandas.Series` of :class:`matplotlib.axes.Axes`
            Pandas array of axes objects

        """
        # TODO: remove this in version >= 2.1
        if 'plot_type' in kwargs:
            raise ValueError(
                "You are using the anesthetic 1.0 kwarg \'plot_type\' instead "
                "of anesthetic 2.0 \'kind\'. Please update your code."
                )

        if not isinstance(axes, AxesSeries):
            _, axes = make_1d_axes(axes, labels=self.get_labels_map())

        kwargs['kind'] = kwargs.get('kind', 'kde_1d')
        kwargs['label'] = kwargs.get('label', self.label)

        # TODO: remove this in version >= 2.1
        if kwargs['kind'] == 'kde':
            warnings.warn(
                "You are using \'kde\' as a plot kind. "
                "\'kde_1d\' is the appropriate keyword for anesthetic. "
                "Your plots may look odd if you use this argument."
                )
        elif kwargs['kind'] == 'hist':
            warnings.warn(
                "You are using \'hist\' as a plot kind. "
                "\'hist_1d\' is the appropriate keyword for anesthetic. "
                "Your plots may look odd if you use this argument."
                )

        for x, ax in axes.items():
            if x in self and kwargs['kind'] is not None:
                xlabel = self.get_label(x)
                self[x].plot(ax=ax, xlabel=xlabel,
                             *args, **kwargs)
                ax.set_xlabel(xlabel)
            else:
                ax.plot([], [])

        return axes

    def plot_2d(self, axes, *args, **kwargs):
        """Create an array of 2D plots.

        To avoid interfering with y-axis sharing, one-dimensional plots are
        created on a separate axis, which is monkey-patched onto the argument
        ax as the attribute ax.twin.

        Parameters
        ----------
        axes : plotting axes
            Can be:
                - list(str) if the x and y axes are the same
                - [list(str),list(str)] if the x and y axes are different
                - :class:`pandas.DataFrame` of :class:`matplotlib.axes.Axes`

            If a :class:`pandas.DataFrame` is provided as an existing set of
            axes, then this is used for creating the plot. Otherwise, a new set
            of axes are created using the list or lists of strings.

        kind/kinds : dict, optional
            What kinds of plots to produce. Dictionary takes the keys
            'diagonal' for the 1D plots and 'lower' and 'upper' for the 2D
            plots. The options for 'diagonal' are:

                - 'kde_1d': :func:`anesthetic.plot.kde_plot_1d`
                - 'hist_1d': :func:`anesthetic.plot.hist_plot_1d`
                - 'fastkde_1d': :func:`anesthetic.plot.fastkde_plot_1d`
                - 'kde': :meth:`pandas.Series.plot.kde`
                - 'hist': :meth:`pandas.Series.plot.hist`
                - 'box': :meth:`pandas.Series.plot.box`
                - 'density': :meth:`pandas.Series.plot.density`

            The options for 'lower' and 'upper' are:

                - 'kde_2d': :func:`anesthetic.plot.kde_contour_plot_2d`
                - 'hist_2d': :func:`anesthetic.plot.hist_plot_2d`
                - 'scatter_2d': :func:`anesthetic.plot.scatter_plot_2d`
                - 'fastkde_2d': :func:`anesthetic.plot.fastkde_contour_plot_2d`
                - 'kde': :meth:`pandas.DataFrame.plot.kde`
                - 'scatter': :meth:`pandas.DataFrame.plot.scatter`
                - 'hexbin': :meth:`pandas.DataFrame.plot.hexbin`

            There are also a set of shortcuts provided in
            :attr:`plot_2d_default_kinds`:

                - 'kde_1d': 1d kde plots down the diagonal
                - 'kde_2d': 2d kde plots in lower triangle
                - 'kde': 1d & 2d kde plots in lower & diagonal
                - 'hist_1d': 1d histograms down the diagonal
                - 'hist_2d': 2d histograms in lower triangle
                - 'hist': 1d & 2d histograms in lower & diagonal

            Feel free to add your own to this list!
            Default:
            {'diagonal': 'kde_1d', 'lower': 'kde_2d', 'upper':'scatter_2d'}

        diagonal_kwargs, lower_kwargs, upper_kwargs : dict, optional
            kwargs for the diagonal (1D)/lower or upper (2D) plots. This is
            useful when there is a conflict of kwargs for different kinds of
            plots.  Note that any kwargs directly passed to plot_2d will
            overwrite any kwarg with the same key passed to <sub>_kwargs.
            Default: {}

        Returns
        -------
        axes : :class:`pandas.DataFrame` of :class:`matplotlib.axes.Axes`
            Pandas array of axes objects

        """
        # TODO: remove this in version >= 2.1
        if 'types' in kwargs:
            raise ValueError(
                "You are using the anesthetic 1.0 kwarg \'types\' instead of "
                "anesthetic 2.0 \'kind' or \'kinds\' (synonyms). "
                "Please update your code."
                )
        kind = kwargs.pop('kind', 'default')
        kind = kwargs.pop('kinds', kind)

        if isinstance(kind, str) and kind in self.plot_2d_default_kinds:
            kind = self.plot_2d_default_kinds.get(kind)
        if (not isinstance(kind, dict) or
                not set(kind.keys()) <= {'lower', 'upper', 'diagonal'}):
            raise ValueError(f"{kind} is not a valid input. `kind`/`kinds` "
                             "must be a dict mapping "
                             "{'lower','diagonal','upper'} to an allowed plot "
                             "(see `help(NestedSamples.plot2d)`), or one of "
                             "the following string shortcuts: "
                             f"{list(self.plot_2d_default_kinds.keys())}")

        local_kwargs = {pos: kwargs.pop('%s_kwargs' % pos, {})
                        for pos in ['upper', 'lower', 'diagonal']}
        kwargs['label'] = kwargs.get('label', self.label)

        for pos in local_kwargs:
            local_kwargs[pos].update(kwargs)

        if not isinstance(axes, AxesDataFrame):
            _, axes = make_2d_axes(axes, labels=self.get_labels(),
                                   upper=('upper' in kind),
                                   lower=('lower' in kind),
                                   diagonal=('diagonal' in kind))

        for y, row in axes.iterrows():
            for x, ax in row.items():
                if ax is not None:
                    pos = ax.position
                    lkwargs = local_kwargs.get(pos, {})
                    lkwargs['kind'] = kind.get(pos, None)
                    # TODO: remove this in version >= 2.1
                    if lkwargs['kind'] == 'kde':
                        warnings.warn(
                            "You are using \'kde\' as a plot kind. "
                            "\'kde_1d\' and \'kde_2d\' are the appropriate "
                            "keywords for anesthetic. Your plots may look "
                            "odd if you use this argument."
                            )
                    elif lkwargs['kind'] == 'hist':
                        warnings.warn(
                            "You are using \'hist\' as a plot kind. "
                            "\'hist_1d\' and \'hist_2d\' are the appropriate "
                            "keywords for anesthetic. Your plots may look "
                            "odd if you use this argument."
                            )
                    if x in self and y in self and lkwargs['kind'] is not None:
                        xlabel = self.get_label(x)
                        ylabel = self.get_label(y)
                        if x == y:
                            self[x].plot(ax=ax.twin, xlabel=xlabel,
                                         *args, **lkwargs)
                            ax.set_xlabel(xlabel)
                            ax.set_ylabel(ylabel)
                        else:
                            self.plot(x, y, ax=ax, xlabel=xlabel,
                                      ylabel=ylabel, *args, **lkwargs)
                            ax.set_xlabel(xlabel)
                            ax.set_ylabel(ylabel)
                    else:
                        if x == y:
                            ax.twin.plot([], [])
                        else:
                            ax.plot([], [])

        return axes

    plot_2d_default_kinds = {
        'default': {'diagonal': 'kde_1d',
                    'lower': 'kde_2d',
                    'upper': 'scatter_2d'},
        'kde': {'diagonal': 'kde_1d', 'lower': 'kde_2d'},
        'kde_1d': {'diagonal': 'kde_1d'},
        'kde_2d': {'lower': 'kde_2d'},
        'hist': {'diagonal': 'hist_1d', 'lower': 'hist_2d'},
        'hist_1d': {'diagonal': 'hist_1d'},
        'hist_2d': {'lower': 'hist_2d'},
    }

    def importance_sample(self, logL_new, action='add', inplace=False):
        """Perform importance re-weighting on the log-likelihood.

        Parameters
        ----------
        logL_new : np.array
            New log-likelihood values. Should have the same shape as `logL`.

        action : str, default='add'
            Can be any of {'add', 'replace', 'mask'}.

            * add: Add the new `logL_new` to the current `logL`.
            * replace: Replace the current `logL` with the new `logL_new`.
            * mask: treat `logL_new` as a boolean mask and only keep the
              corresponding (True) samples.

        inplace : bool, default=False
            Indicates whether to modify the existing array, or return a new
            frame with importance sampling applied.

        Returns
        -------
        samples : :class:`Samples`/:class:`MCMCSamples`/:class:`NestedSamples`
            Importance re-weighted samples.

        """
        if inplace:
            samples = self
        else:
            samples = self.copy()

        if action == 'add':
            new_weights = samples.get_weights()
            new_weights *= np.exp(logL_new - logL_new.max())
            samples.set_weights(new_weights, inplace=True)
            samples.logL += logL_new
        elif action == 'replace':
            logL_new2 = logL_new - samples.logL
            new_weights = samples.get_weights()
            new_weights *= np.exp(logL_new2 - logL_new2.max())
            samples.set_weights(new_weights, inplace=True)
            samples.logL = logL_new
        elif action == 'mask':
            samples = samples[logL_new]
        else:
            raise NotImplementedError("`action` needs to be one of "
                                      "{'add', 'replace', 'mask'}, but '%s' "
                                      "was requested." % action)

        if inplace:
            self._update_inplace(samples)
        else:
            return samples.__finalize__(self, "importance_sample")

    # TODO: remove this in version >= 2.1
    @property
    def tex(self):
        # noqa: disable=D102
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "samples.tex[label] = tex        # anesthetic 1.0\n"
            "samples.set_label(label, tex)   # anesthetic 2.0\n\n"
            "tex = samples.tex[label]        # anesthetic 1.0\n"
            "tex = samples.get_label(label)  # anesthetic 2.0"
            )


class MCMCSamples(Samples):
    """Storage and plotting tools for MCMC samples.

    Any new functionality specific to MCMC (e.g. convergence criteria etc.)
    should be put here.

    Parameters
    ----------
    data : np.array
        Coordinates of samples. shape = (nsamples, ndims).

    columns : array-like
        reference names of parameters

    weights : np.array
        weights of samples.

    logL : np.array
        loglikelihoods of samples.

    labels : dict or array-like
        mapping from columns to plotting labels

    label : str
        Legend label

    logzero : float, default=-1e30
        The threshold for `log(0)` values assigned to rejected sample points.
        Anything equal or below this value is set to `-np.inf`.

    """

    _metadata = Samples._metadata + ['root']

    @property
    def _constructor(self):
        return MCMCSamples

    def remove_burn_in(self, burn_in, reset_index=False, inplace=False):
        """Remove burn-in samples from each MCMC chain.

        Parameters
        ----------
        burn_in : int or float or array_like
            Fraction or number of samples to remove or keep:

            * ``if 0 < burn_in < 1``: remove first fraction of samples
            * ``elif 1 < burn_in``: remove first number of samples
            * ``elif -1 < burn_in < 0``: keep last fraction of samples
            * ``elif burn_in < -1``: keep last number of samples
            * ``elif type(burn_in)==list``: different burn-in for each chain

        reset_index : bool, default=False
            Whether to reset the index counter to start at zero or not.

        inplace : bool, default=False
            Indicates whether to modify the existing array or return a copy.

        """
        chains = self.groupby(('chain', '$n_\\mathrm{chain}$'), sort=False,
                              group_keys=False)
        nchains = chains.ngroups
        if isinstance(burn_in, (int, float)):
            ndrop = np.full(nchains, burn_in)
        elif isinstance(burn_in, (list, tuple, np.ndarray)) \
                and len(burn_in) == nchains:
            ndrop = np.array(burn_in)
        else:
            raise ValueError("`burn_in` has to be a scalar or an array of "
                             "length matching the number of chains "
                             "`nchains=%d`. However, you provided "
                             "`burn_in=%s`" % (nchains, burn_in))
        if np.all(np.abs(ndrop) < 1):
            nsamples = chains.count().iloc[:, 0].to_numpy()
            ndrop = ndrop * nsamples
        ndrop = ndrop.astype(int)
        data = self.drop(chains.apply(lambda g: g.head(ndrop[g.name-1])).index,
                         inplace=inplace)
        if reset_index:
            data = data.reset_index(drop=True, inplace=inplace)
        return data

    def Gelman_Rubin(self, params=None):
        """Gelman--Rubin convergence statistic of multiple MCMC chains.

        Determine the Gelman--Rubin convergence statistic ``R-1`` by computing
        and comparing the within-chain variance and the between-chain variance.
        This follows the routine as outlined in
        `Lewis (2013), section IV.A. <https://arxiv.org/abs/1304.4473>`_

        Note that this requires more than one chain. To circumvent this, you
        could overwrite the ``'chain'`` column, splitting the samples into two
        or more sets.

        Parameters
        ----------
        params : list(str)
            List of column names (i.e. parameters) to be included in the
            convergence calculation.
            Default: all parameters (except those parameters that contain
            'prior', 'chi2', or 'logL' in their names)

        Returns
        -------
        Rminus1 : float
            Gelman--Rubin convergence statistic ``R-1``. The smaller, the
            better converged. Aiming for ``Rminus1~0.01`` should normally work
            well.

        """
        self.columns.set_names(['params', 'labels'], inplace=True)
        if params is None:
            params = [key for key in self.columns.get_level_values('params')
                      if 'prior' not in key
                      and 'chi2' not in key
                      and 'logL' not in key
                      and 'chain' not in key]
        chains = self[params+['chain']].groupby(
                ('chain', '$n_\\mathrm{chain}$'), sort=False,
        )
        nchains = chains.ngroups

        # Within chain variance ``W``
        # (average variance within each chain):
        W = chains.cov().groupby(level=('params', 'labels'), sort=False).mean()
        # Between-chain variance ``B``
        # (variance of the chain means):
        B = np.atleast_2d(np.cov(chains.mean().T, ddof=1))
        # We don't weight `B` with the effective number of samples (sum of the
        # weights), here, because we want to notice outliers from shorter
        # chains.
        # In order to be conservative, we generally want to underestimate `W`
        # and overestimate `B`, since `W` goes in the denominator and `B` in
        # the numerator of the Gelman--Rubin statistic `Rminus1`.

        try:
            invL = np.linalg.inv(scipy.linalg.cholesky(W))
        except np.linalg.LinAlgError as e:
            raise np.linalg.LinAlgError(
                "Make sure you do not have linearly dependent parameters, "
                "e.g. having both `As` and `A=1e9*As` causes trouble.") from e
        D = np.linalg.eigvalsh(invL @ ((nchains+1)/nchains * B) @ invL.T)
        # The factor of `(nchains+1)/nchains` accounts for the additional
        # uncertainty from using a finite number of chains.
        Rminus1 = np.max(np.abs(D))
        return Rminus1


class NestedSamples(Samples):
    """Storage and plotting tools for Nested Sampling samples.

    We extend the :class:`Samples` class with the additional methods:

    * ``self.live_points(logL)``
    * ``self.set_beta(beta)``
    * ``self.prior()``
    * ``self.posterior_points(beta)``
    * ``self.prior_points()``
    * ``self.stats()``
    * ``self.logZ()``
    * ``self.D_KL()``
    * ``self.d()``
    * ``self.recompute()``
    * ``self.gui()``
    * ``self.importance_sample()``

    Parameters
    ----------
    data : np.array
        Coordinates of samples. shape = (nsamples, ndims).

    columns : list(str)
        reference names of parameters

    logL : np.array
        loglikelihoods of samples.

    logL_birth : np.array or int
        birth loglikelihoods, or number of live points.

    labels : dict
        optional mapping from column names to plot labels

    label : str
        Legend label
        default: basename of root

    beta : float
        thermodynamic temperature
        default: 1.

    logzero : float
        The threshold for `log(0)` values assigned to rejected sample points.
        Anything equal or below this value is set to `-np.inf`.
        default: -1e30

    """

    _metadata = Samples._metadata + ['root', '_beta']

    def __init__(self, *args, **kwargs):
        logzero = kwargs.pop('logzero', -1e30)
        self._beta = kwargs.pop('beta', 1.)
        logL_birth = kwargs.pop('logL_birth', None)
        if not isinstance(logL_birth, int) and logL_birth is not None:
            logL_birth = np.array(logL_birth)
            logL_birth = np.where(logL_birth <= logzero, -np.inf,
                                  logL_birth)

        super().__init__(logzero=logzero, *args, **kwargs)
        if logL_birth is not None:
            self.recompute(logL_birth, inplace=True)

    @property
    def _constructor(self):
        return NestedSamples

    def _compute_insertion_indexes(self):
        logL = self.logL.to_numpy()
        logL_birth = self.logL_birth.to_numpy()
        self['insertion'] = compute_insertion_indexes(logL, logL_birth)

    @property
    def beta(self):
        """Thermodynamic inverse temperature."""
        return self._beta

    @beta.setter
    def beta(self, beta):
        self._beta = beta
        logw = self.logw(beta=beta)
        self.set_weights(np.exp(logw - logw.max()), inplace=True)

    def set_beta(self, beta, inplace=False):
        """Change the inverse temperature.

        Parameters
        ----------
        beta : float
            Temperature to set.
            (``beta=0`` corresponds to the prior distribution.)

        inplace : bool, default=False
            Indicates whether to modify the existing array, or return a copy
            with the temperature changed.

        """
        if inplace:
            self.beta = beta
        else:
            data = self.copy()
            data.beta = beta
            return data

    def prior(self, inplace=False):
        """Re-weight samples at infinite temperature to get prior samples."""
        return self.set_beta(beta=0, inplace=inplace)

    # TODO: remove this in version >= 2.1
    def ns_output(self, *args, **kwargs):
        # noqa: disable=D102
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "samples.ns_output(1000)  # anesthetic 1.0\n"
            "samples.stats(1000)      # anesthetic 2.0\n\n"
            "Check out the new temperature functionality: help(samples.stats),"
            " as well as average loglikelihoods: help(samples.logL_P)"
            )

    def stats(self, nsamples=None, beta=None):
        r"""Compute Nested Sampling statistics.

        Using nested sampling we can compute:

        - ``logZ``: Bayesian evidence

          .. math::
              \log Z = \int L \pi d\theta

        - ``D_KL``: Kullback--Leibler divergence

          .. math::
              D_{KL} = \int P \log(P / \pi) d\theta

        - ``logL_P``: posterior averaged log-likelihood

          .. math::
              \langle\log L\rangle_P = \int P \log L d\theta

        - ``d_G``: Gaussian model dimensionality
          (or posterior variance of the log-likelihood)

          .. math::
              d_G/2 = \langle(\log L)^2\rangle_P - \langle\log L\rangle_P^2

          see `Handley and Lemos (2019) <https://arxiv.org/abs/1903.06682>`_
          for more details on model dimensionalities.

        (Note that all of these are available as individual functions with the
        same signature.)

        In addition to point estimates nested sampling provides an error bar
        or more generally samples from a (correlated) distribution over the
        variables. Samples from this distribution can be computed by providing
        an integer nsamples.

        Nested sampling as an athermal algorithm is also capable of producing
        these as a function of inverse thermodynamic temperature beta. This is
        provided as a vectorised function. If nsamples is also provided a
        MultiIndex dataframe is generated.

        These obey Occam's razor equation:

        .. math::
            \log Z = \langle\log L\rangle_P - D_{KL},

        which splits a model's quality ``logZ`` into a goodness-of-fit
        ``logL_P`` and a complexity penalty ``D_KL``. See `Hergt et al. (2021)
        <https://arxiv.org/abs/2102.11511>`_ for more detail.

        Parameters
        ----------
        nsamples : int, optional
            - If nsamples is not supplied, calculate mean value
            - If nsamples is integer, draw nsamples from the distribution of
              values inferred by nested sampling

        beta : float, array-like, optional
            inverse temperature(s) beta=1/kT. Default self.beta

        Returns
        -------
        if beta is scalar and nsamples is None:
            Series, index ['logZ', 'd_G', 'DK_L', 'logL_P']
        elif beta is scalar and nsamples is int:
            :class:`Samples`, index range(nsamples),
            columns ['logZ', 'd_G', 'DK_L', 'logL_P']
        elif beta is array-like and nsamples is None:
            :class:`Samples`, index beta,
            columns ['logZ', 'd_G', 'DK_L', 'logL_P']
        elif beta is array-like and nsamples is int:
            :class:`Samples`, index :class:`pandas.MultiIndex` the product of
            beta and range(nsamples)
            columns ['logZ', 'd_G', 'DK_L', 'logL_P']
        """
        logw = self.logw(nsamples, beta)
        if nsamples is None and beta is None:
            samples = self._constructor_sliced(index=self.columns[:0],
                                               dtype=float)
        else:
            samples = Samples(index=logw.columns, columns=self.columns[:0])
        samples['logZ'] = self.logZ(logw)
        samples.set_label('logZ', r'$\ln\mathcal{Z}$')
        w = np.exp(logw-samples['logZ'])

        betalogL = self._betalogL(beta)
        S = (logw*0).add(betalogL, axis=0) - samples.logZ

        samples['D_KL'] = (S*w).sum()
        samples.set_label('D_KL', r'$\mathcal{D}_\mathrm{KL}$')

        samples['logL_P'] = samples['logZ'] + samples['D_KL']
        samples.set_label('logL_P',
                          r'$\langle\ln\mathcal{L}\rangle_\mathcal{P}$')

        samples['d_G'] = ((S-samples.D_KL)**2*w).sum()*2
        samples.set_label('d_G', r'$d_\mathrm{G}$')

        samples.label = self.label
        return samples

    def logX(self, nsamples=None):
        """Log-Volume.

        The log of the prior volume contained within each iso-likelihood
        contour.

        Parameters
        ----------
        nsamples : int, optional
            - If nsamples is not supplied, calculate mean value
            - If nsamples is integer, draw nsamples from the distribution of
              values inferred by nested sampling

        Returns
        -------
        if nsamples is None:
            WeightedSeries like self
        elif nsamples is int:
            WeightedDataFrame like self, columns range(nsamples)
        """
        if nsamples is None:
            t = np.log(self.nlive/(self.nlive+1))
        else:
            r = np.log(np.random.rand(len(self), nsamples))
            w = self.get_weights()
            r = self.nlive._constructor_expanddim(r, self.index, weights=w)
            t = r.divide(self.nlive, axis=0)
            t.columns.name = 'samples'
        logX = t.cumsum()
        logX.name = 'logX'
        return logX

    def dlogX(self, nsamples=None):
        """Compute volume of shell of loglikelihood.

        Parameters
        ----------
        nsamples : int, optional
            - If nsamples is not supplied, calculate mean value
            - If nsamples is integer, draw nsamples from the distribution of
              values inferred by nested sampling

        Returns
        -------
        if nsamples is None:
            WeightedSeries like self
        elif nsamples is int:
            WeightedDataFrame like self, columns range(nsamples)
        """
        logX = self.logX(nsamples)
        logXp = logX.shift(1, fill_value=0)
        logXm = logX.shift(-1, fill_value=-np.inf)
        dlogX = np.log(1 - np.exp(logXm-logXp)) + logXp - np.log(2)
        dlogX.name = 'dlogX'

        return dlogX

    def _betalogL(self, beta=None):
        """Log(L**beta) convenience function.

        Parameters
        ----------
        beta : scalar or array-like, optional
            inverse temperature(s) beta=1/kT. Default self.beta

        Returns
        -------
        if beta is scalar:
            WeightedSeries like self
        elif beta is array-like:
            WeightedDataFrame like self, columns of beta
        """
        if beta is None:
            beta = self.beta
        logL = self.logL
        if np.isscalar(beta):
            betalogL = beta * logL
            betalogL.name = 'betalogL'
        else:
            betalogL = logL._constructor_expanddim(np.outer(self.logL, beta),
                                                   self.index, columns=beta)
            betalogL.columns.name = 'beta'
        return betalogL

    def logw(self, nsamples=None, beta=None):
        """Log-nested sampling weight.

        The logarithm of the (unnormalised) sampling weight log(L**beta*dX).

        Parameters
        ----------
        nsamples : int, optional
            - If nsamples is not supplied, calculate mean value
            - If nsamples is integer, draw nsamples from the distribution of
              values inferred by nested sampling
            - If nsamples is array, nsamples is assumed to be logw and returned
              (implementation convenience functionality)

        beta : float, array-like, optional
            inverse temperature(s) beta=1/kT. Default self.beta

        Returns
        -------
        if nsamples is array-like:
            WeightedDataFrame equal to nsamples
        elif beta is scalar and nsamples is None:
            WeightedSeries like self
        elif beta is array-like and nsamples is None:
            WeightedDataFrame like self, columns of beta
        elif beta is scalar and nsamples is int:
            WeightedDataFrame like self, columns of range(nsamples)
        elif beta is array-like and nsamples is int:
            WeightedDataFrame like self, MultiIndex columns the product of beta
            and range(nsamples)
        """
        if np.ndim(nsamples) > 0:
            return nsamples

        dlogX = self.dlogX(nsamples)
        betalogL = self._betalogL(beta)

        if dlogX.ndim == 1 and betalogL.ndim == 1:
            logw = dlogX + betalogL
        elif dlogX.ndim > 1 and betalogL.ndim == 1:
            logw = dlogX.add(betalogL, axis=0)
        elif dlogX.ndim == 1 and betalogL.ndim > 1:
            logw = betalogL.add(dlogX, axis=0)
        else:
            cols = MultiIndex.from_product([betalogL.columns, dlogX.columns])
            dlogX = dlogX.reindex(columns=cols, level='samples')
            betalogL = betalogL.reindex(columns=cols, level='beta')
            logw = betalogL+dlogX
        return logw

    def logZ(self, nsamples=None, beta=None):
        """Log-Evidence.

        Parameters
        ----------
        nsamples : int, optional
            - If nsamples is not supplied, calculate mean value
            - If nsamples is integer, draw nsamples from the distribution of
              values inferred by nested sampling
            - If nsamples is array, nsamples is assumed to be logw

        beta : float, array-like, optional
            inverse temperature(s) beta=1/kT. Default self.beta

        Returns
        -------
        if nsamples is array-like:
            :class:`pandas.Series`, index nsamples.columns
        elif beta is scalar and nsamples is None:
            float
        elif beta is array-like and nsamples is None:
            :class:`pandas.Series`, index beta
        elif beta is scalar and nsamples is int:
            :class:`pandas.Series`, index range(nsamples)
        elif beta is array-like and nsamples is int:
            :class:`pandas.Series`, :class:`pandas.MultiIndex` columns the
            product of beta and range(nsamples)
        """
        logw = self.logw(nsamples, beta)
        logZ = logsumexp(logw, axis=0)
        if np.isscalar(logZ):
            return logZ
        else:
            return logw._constructor_sliced(logZ, name='logZ',
                                            index=logw.columns).squeeze()

    _logZ_function_shape = '\n' + '\n'.join(logZ.__doc__.split('\n')[1:])

    # TODO: remove this in version >= 2.1
    def D(self, nsamples=None):
        # noqa: disable=D102
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "samples.D(1000)     # anesthetic 1.0\n"
            "samples.D_KL(1000)  # anesthetic 2.0\n\n"
            "Check out the new temperature functionality: help(samples.D_KL), "
            "as well as average loglikelihoods: help(samples.logL_P)"
            )

    def D_KL(self, nsamples=None, beta=None):
        """Kullback--Leibler divergence."""
        logw = self.logw(nsamples, beta)
        logZ = self.logZ(logw, beta)
        betalogL = self._betalogL(beta)
        S = (logw*0).add(betalogL, axis=0) - logZ
        w = np.exp(logw-logZ)
        D_KL = (S*w).sum()
        if np.isscalar(D_KL):
            return D_KL
        else:
            return self._constructor_sliced(D_KL, name='D_KL',
                                            index=logw.columns).squeeze()

    D_KL.__doc__ += _logZ_function_shape

    # TODO: remove this in version >= 2.1
    def d(self, nsamples=None):
        # noqa: disable=D102
        raise NotImplementedError(
            "This is anesthetic 1.0 syntax. You need to update, e.g.\n"
            "samples.d(1000)     # anesthetic 1.0\n"
            "samples.d_G(1000)  # anesthetic 2.0\n\n"
            "Check out the new temperature functionality: help(samples.d_G), "
            "as well as average loglikelihoods: help(samples.logL_P)"
            )

    def d_G(self, nsamples=None, beta=None):
        """Bayesian model dimensionality."""
        logw = self.logw(nsamples, beta)
        logZ = self.logZ(logw, beta)
        betalogL = self._betalogL(beta)
        S = (logw*0).add(betalogL, axis=0) - logZ
        w = np.exp(logw-logZ)
        D_KL = (S*w).sum()
        d_G = ((S-D_KL)**2*w).sum()*2
        if np.isscalar(d_G):
            return d_G
        else:
            return self._constructor_sliced(d_G, name='d_G',
                                            index=logw.columns).squeeze()

    d_G.__doc__ += _logZ_function_shape

    def logL_P(self, nsamples=None, beta=None):
        """Posterior averaged loglikelihood."""
        logw = self.logw(nsamples, beta)
        logZ = self.logZ(logw, beta)
        betalogL = self._betalogL(beta)
        betalogL = (logw*0).add(betalogL, axis=0)
        w = np.exp(logw-logZ)
        logL_P = (betalogL*w).sum()
        if np.isscalar(logL_P):
            return logL_P
        else:
            return self._constructor_sliced(logL_P, name='logL_P',
                                            index=logw.columns).squeeze()

    logL_P.__doc__ += _logZ_function_shape

    def live_points(self, logL=None):
        """Get the live points within logL.

        Parameters
        ----------
        logL : float or int, optional
            Loglikelihood or iteration number to return live points.
            If not provided, return the last set of active live points.

        Returns
        -------
        live_points : Samples
            Live points at either:
                - contour logL (if input is float)
                - ith iteration (if input is integer)
                - last set of live points if no argument provided
        """
        if logL is None:
            logL = self.logL_birth.max()
        else:
            try:
                logL = float(self.logL[logL])
            except KeyError:
                pass
        i = ((self.logL >= logL) & (self.logL_birth < logL)).to_numpy()
        return Samples(self[i]).set_weights(None)

    def posterior_points(self, beta=1):
        """Get equally weighted posterior points at temperature beta."""
        return self.set_beta(beta).compress(-1)

    def prior_points(self, params=None):
        """Get equally weighted prior points."""
        return self.posterior_points(beta=0)

    def gui(self, params=None):
        """Construct a graphical user interface for viewing samples."""
        return RunPlotter(self, params)

    def importance_sample(self, logL_new, action='add', inplace=False):
        """Perform importance re-weighting on the log-likelihood.

        Parameters
        ----------
        logL_new : np.array
            New log-likelihood values. Should have the same shape as `logL`.

        action : str, default='add'
            Can be any of {'add', 'replace', 'mask'}.

            * add: Add the new `logL_new` to the current `logL`.
            * replace: Replace the current `logL` with the new `logL_new`.
            * mask: treat `logL_new` as a boolean mask and only keep the
              corresponding (True) samples.

        inplace : bool, optional
            Indicates whether to modify the existing array, or return a new
            frame with importance sampling applied.
            default: False

        Returns
        -------
        samples : :class:`NestedSamples`
            Importance re-weighted samples.

        """
        samples = super().importance_sample(logL_new, action=action)
        mask = (samples.logL > samples.logL_birth).to_numpy()
        samples = samples[mask].recompute()
        if inplace:
            self._update_inplace(samples)
        else:
            return samples.__finalize__(self, "importance_sample")

    def recompute(self, logL_birth=None, inplace=False):
        """Re-calculate the nested sampling contours and live points.

        Parameters
        ----------
        logL_birth : array-like or int, optional

            * array-like: the birth contours.
            * int: the number of live points.
            * default: use the existing birth contours to compute nlive

        inplace : bool, default=False
            Indicates whether to modify the existing array, or return a new
            frame with contours resorted and nlive recomputed

        """
        if inplace:
            samples = self
        else:
            samples = self.copy()

        nlive_label = r'$n_\mathrm{live}$'
        if is_int(logL_birth):
            nlive = logL_birth
            samples.sort_values('logL', inplace=True)
            samples.reset_index(drop=True, inplace=True)
            n = np.ones(len(self), int) * nlive
            n[-nlive:] = np.arange(nlive, 0, -1)
            samples['nlive', nlive_label] = n
        else:
            if logL_birth is not None:
                label = r'$\ln\mathcal{L}_\mathrm{birth}$'
                samples['logL_birth'] = logL_birth
                if self.islabelled():
                    samples.set_label('logL_birth', label)

            if 'logL_birth' not in samples:
                raise RuntimeError("Cannot recompute run without "
                                   "birth contours logL_birth.")

            invalid = (samples.logL <= samples.logL_birth).to_numpy()
            n_bad = invalid.sum()
            n_equal = (samples.logL == samples.logL_birth).sum()
            if n_bad:
                warnings.warn("%i out of %i samples have logL <= logL_birth,"
                              "\n%i of which have logL == logL_birth."
                              "\nThis may just indicate numerical rounding "
                              "errors at the peak of the likelihood, but "
                              "further investigation of the chains files is "
                              "recommended."
                              "\nDropping the invalid samples." %
                              (n_bad, len(samples), n_equal),
                              RuntimeWarning)
                samples = samples[~invalid].reset_index(drop=True)

            samples.sort_values('logL', inplace=True)
            samples.reset_index(drop=True, inplace=True)
            nlive = compute_nlive(samples.logL, samples.logL_birth)
            samples['nlive'] = nlive
            if self.islabelled():
                samples.set_label('nlive', nlive_label)

        samples.beta = samples._beta

        if np.any(pandas.isna(samples.logL)):
            warnings.warn("NaN encountered in logL. If this is unexpected, you"
                          " should investigate why your likelihood is throwing"
                          " NaNs. Dropping these samples at prior level",
                          RuntimeWarning)
            samples = samples[samples.logL.notna().to_numpy()].recompute()

        if inplace:
            self._update_inplace(samples)
        else:
            return samples.__finalize__(self, "recompute")


def merge_nested_samples(runs):
    """Merge one or more nested sampling runs.

    Parameters
    ----------
    runs : list(:class:`NestedSamples`)
        List or array-like of one or more nested sampling runs.
        If only a single run is provided, this recalculates the live points and
        as such can be used for masked runs.

    Returns
    -------
    samples : :class:`NestedSamples`
        Merged run.
    """
    merge = pandas.concat(runs, ignore_index=True)
    return merge.recompute()


def merge_samples_weighted(samples, weights=None, label=None):
    r"""Merge sets of samples with weights.

    Combine two (or more) samples so the new PDF is
    P(x|new) = weight_A P(x|A) + weight_B P(x|B).
    The number of samples and internal weights do not affect the result.

    Parameters
    ----------
    samples : list(:class:`NestedSamples`) or list(:class:`MCMCSamples`)
        List or array-like of one or more MCMC or nested sampling runs.

    weights : list(double) or None
        Weight for each run in samples (normalized internally).
        Can be omitted if samples are :class:`NestedSamples`,
        then exp(logZ) is used as weight.

    label : str or None, default=None
        Label for the new samples.

    Returns
    -------
    new_samples : :class:`Samples`
        Merged (weighted) run.
    """
    if not (isinstance(samples, Sequence) or
            isinstance(samples, Series)):
        raise TypeError("samples must be a list of samples "
                        "(Sequence or pandas.Series)")

    mcmc_samples = copy.deepcopy([Samples(s) for s in samples])
    if weights is None:
        try:
            logZs = np.array(copy.deepcopy([s.logZ() for s in samples]))
        except AttributeError:
            raise ValueError("If samples includes MCMCSamples "
                             "then weights must be given.")
        # Subtract logsumexp to avoid numerical issues (similar to max(logZs))
        logZs -= logsumexp(logZs)
        weights = np.exp(logZs)
    else:
        if len(weights) != len(samples):
            raise ValueError("samples and weights must have the same length,"
                             "each weight is for a whole sample. Currently",
                             len(samples), len(weights))

    new_samples = []
    for s, w in zip(mcmc_samples, weights):
        # Normalize the given weights
        new_weights = s.get_weights() / s.get_weights().sum()
        new_weights *= w/np.sum(weights)
        s = Samples(s, weights=new_weights)
        new_samples.append(s)

    new_samples = pandas.concat(new_samples)

    new_weights = new_samples.get_weights()
    new_weights /= new_weights.max()
    new_samples.set_weights(new_weights, inplace=True)

    new_samples.label = label

    return new_samples



"""Read NestedSamples from polychord chains."""
import os
import numpy as np
from anesthetic.read.getdist import read_paramnames
from anesthetic.samples import NestedSamples


def read_polychord(root, *args, **kwargs):
    """Read ``<root>_dead-birth.txt`` in PolyChord format."""
    birth_file = root + '_dead-birth.txt'
    birth_file
    data = np.loadtxt(birth_file)
    try:
        phys_live_birth_file = root + '_phys_live-birth.txt'
        _data = np.loadtxt(phys_live_birth_file)
        _data = np.atleast_2d(_data)
        data = np.concatenate([data, _data]) if _data.size else data
        data = np.unique(data, axis=0)
        i = np.argsort(data[:, -2])
        data = data[i, :]
    except IOError:
        pass
    data, logL, logL_birth = np.split(data, [-2, -1], axis=1)
    columns, labels = read_paramnames(root)

    columns = kwargs.pop('columns', columns)
    labels = kwargs.pop('labels', labels)
    kwargs['label'] = kwargs.get('label', os.path.basename(root))

    return NestedSamples(data=data, columns=columns,
                         logL=logL, logL_birth=logL_birth,
                         labels=labels, *args, **kwargs)



import warnings
import numpy as np
import pytest
from scipy import special as sp
from numpy.testing import assert_array_equal
from anesthetic import read_chains
from anesthetic.utils import (nest_level, compute_nlive, unique, is_int,
                              iso_probability_contours,
                              iso_probability_contours_from_samples,
                              logsumexp, sample_compression_1d,
                              triangular_sample_compression_2d,
                              insertion_p_value, compress_weights)


def test_compress_weights():
    w = compress_weights(w=np.ones(10), u=None)
    assert_array_equal(w, np.ones(10))
    w = compress_weights(w=None, u=np.random.rand(10))
    assert_array_equal(w, np.ones(10))
    r = np.random.rand(10)
    w = compress_weights(w=r, u=None, ncompress=False)
    assert_array_equal(w, r)


def test_nest_level():
    assert nest_level(0) == 0
    assert nest_level([]) == 1
    assert nest_level(['a']) == 1
    assert nest_level(['a', 'b']) == 1
    assert nest_level([['a'], 'b']) == 2
    assert nest_level(['a', ['b']]) == 2
    assert nest_level([['a'], ['b']]) == 2


def test_compute_nlive():
    # Generate a 'pure' nested sampling run
    np.random.seed(0)
    nlive = 500
    ncompress = 100
    logL = np.cumsum(np.random.rand(nlive, ncompress), axis=1)
    logL_birth = np.concatenate((np.ones((nlive, 1))*-np.inf, logL[:, :-1]),
                                axis=1)
    i = np.argsort(logL.flatten())
    logL = logL.flatten()[i]
    logL_birth = logL_birth.flatten()[i]

    # Compute nlive
    nlives = compute_nlive(logL, logL_birth)

    # Check the first half are constant
    assert_array_equal(nlives[:len(nlives)//2], nlive)

    # Check one point at the end
    assert nlives[-1] == 1

    # Check never more than nlive
    assert nlives.max() <= nlive

    # Check length
    assert (len(nlives) == len(logL))


def test_unique():
    assert unique([3, 2, 1, 4, 1, 3]) == [3, 2, 1, 4]


@pytest.mark.parametrize('ipc', [iso_probability_contours,
                                 iso_probability_contours_from_samples])
def test_iso_probability_contours(ipc):
    p = np.random.randn(10)
    ipc(p)
    with pytest.raises(ValueError):
        ipc(p, contours=[0.68, 0.95])


def test_triangular_sample_compression_2d():
    np.random.seed(0)
    n = 5000
    x = np.random.rand(n)
    y = np.random.rand(n)
    w = np.random.rand(n)
    cov = np.identity(2)
    tri, W = triangular_sample_compression_2d(x, y, cov, w)
    assert len(W) == 1000
    assert np.isclose(sum(W), sum(w), rtol=1e-1)


def test_sample_compression_1d():
    np.random.seed(0)
    N = 10000
    x_ = np.random.rand(N)
    w_ = np.random.rand(N)
    n = 1000
    x, w = sample_compression_1d(x_, w_, n)
    assert len(x) == n
    assert len(w) == n
    assert np.isclose(w.sum(), w_.sum())

    x_ = np.random.rand(N)
    w_ = np.random.randn(N)
    x, w = sample_compression_1d(x_, w_, ncompress=True)
    assert len(w) <= N


def test_is_int():
    assert is_int(1)
    assert is_int(np.int64(1))
    assert not is_int(1.)
    assert not is_int(np.float64(1.))


def test_logsumexpinf():
    np.random.seed(0)
    a = np.random.rand(10)
    b = np.random.rand(10)
    assert logsumexp(-np.inf, b=[-np.inf]) == -np.inf
    assert logsumexp(a, b=b) == sp.logsumexp(a, b=b)
    a[0] = -np.inf
    assert logsumexp(a, b=b) == sp.logsumexp(a, b=b)
    b[0] = -np.inf
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',
                                'invalid value encountered in multiply',
                                RuntimeWarning)
        assert np.isnan(sp.logsumexp(a, b=b))
    assert np.isfinite(logsumexp(a, b=b))


def test_insertion_p_value():
    np.random.seed(3)
    nlive = 500
    ndead = nlive*20
    indexes = np.random.randint(0, nlive, ndead)
    ks_results = insertion_p_value(indexes, nlive)
    assert 'D' in ks_results
    assert 'p-value' in ks_results
    assert 'sample_size' in ks_results

    assert 'iterations' not in ks_results
    assert 'nbatches' not in ks_results
    assert 'p_value_uncorrected' not in ks_results

    assert ks_results['p-value'] > 0.05
    assert ks_results['sample_size'] == ndead

    ks_results = insertion_p_value(indexes, nlive, 1)
    assert 'D' in ks_results
    assert 'p-value' in ks_results
    assert 'sample_size' in ks_results
    assert 'iterations' in ks_results
    assert 'nbatches' in ks_results
    assert 'uncorrected p-value' in ks_results

    assert ks_results['p-value'] > 0.05
    assert ks_results['uncorrected p-value'] < ks_results['p-value']

    iterations = ks_results['iterations']
    assert isinstance(iterations, tuple)
    assert len(iterations) == 2
    assert iterations[1] - iterations[0] == nlive
    assert ks_results['nbatches'] == 20


def test_p_values_from_sample():
    np.random.seed(3)
    ns = read_chains('./tests/example_data/pc')
    ns._compute_insertion_indexes()
    nlive = len(ns.live_points())

    ks_results = insertion_p_value(ns.insertion[nlive:-nlive], nlive)
    assert ks_results['p-value'] > 0.05

    ks_results = insertion_p_value(ns.insertion[nlive:-nlive], nlive, batch=1)
    assert ks_results['p-value'] > 0.05



"""Boundary correction utilities."""

import numpy as np
from scipy.special import erf


def cut_and_normalise_gaussian(x, p, bw, xmin=None, xmax=None):
    """Cut and normalise boundary correction for a Gaussian kernel.

    Parameters
    ----------
    x : array-like
        locations for normalisation correction

    p : array-like
        probability densities for normalisation correction

    bw : float
        bandwidth of KDE

    xmin, xmax : float
        lower/upper prior bound
        optional, default None

    Returns
    -------
    p : np.array
        corrected probabilities

    """
    correction = np.ones_like(x)

    if xmin is not None:
        correction *= 0.5 * (1 + erf((x-xmin)/bw/np.sqrt(2)))
        correction[x < xmin] = np.inf
    if xmax is not None:
        correction *= 0.5 * (1 + erf((xmax-x)/bw/np.sqrt(2)))
        correction[x > xmax] = np.inf
    return p/correction



"""Read MCMCSamples or NestedSamples from any chains."""
from anesthetic.read.polychord import read_polychord
from anesthetic.read.getdist import read_getdist
from anesthetic.read.cobaya import read_cobaya
from anesthetic.read.multinest import read_multinest


def read_chains(root, *args, **kwargs):
    """Auto-detect chain type and read from file.

    anesthetic supports chains from:

        * `PolyChord <https://github.com/PolyChord/PolyChordLite>`_,
        * `MultiNest <https://github.com/farhanferoz/MultiNest>`_,
        * `CosmoMC <https://github.com/cmbant/CosmoMC>`_,
        * `Cobaya <https://github.com/CobayaSampler/cobaya>`_,
        * or anything `GetDist <https://github.com/cmbant/getdist>`_
          compatible.

    Note that in order to optimally read chains from Cobaya you need to have
    `GetDist <https://getdist.readthedocs.io/en/latest/>`__ installed.

    Parameters
    ----------
    root : str
        root name for reading files

    *args, **kwargs:
        Passed onto ``NestedSamples`` or ``MCMCSamples``. Check their
        docstrings for more information.

    Returns
    -------
    :class:`anesthetic.samples.NestedSamples` or
    :class:`anesthetic.samples.MCMCSamples` depending on auto-detection

    """
    if 'burn_in' in kwargs:
        raise KeyError(
            "This is anesthetic 1.0 syntax. The `burn_in` keyword is no "
            "longer supported in `read_chains`. You need to update, e.g.\n"
            "read_chains(root, burn_in=0.5)         # anesthetic 1.0\n"
            "read_chains(root).remove_burn_in(0.5)  # anesthetic 2.0\n"
            "See also https://anesthetic.readthedocs.io/en/latest/"
            "anesthetic.html#anesthetic.samples.MCMCSamples.remove_burn_in"
        )
    errors = []
    for read in [read_polychord, read_multinest, read_cobaya, read_getdist]:
        try:
            samples = read(root, *args, **kwargs)
            samples.root = root
            return samples
        except (FileNotFoundError, IOError) as e:
            errors.append(str(read) + ": " + str(e))

    errors = ["Could not find any compatible chains:"] + errors
    raise FileNotFoundError('\n'.join(errors))



"""Data-processing utility functions."""
import numpy as np
import pandas
from scipy import special
from scipy.interpolate import interp1d
from scipy.stats import kstwobign
from matplotlib.tri import Triangulation
import contextlib
import inspect
import re


def logsumexp(a, axis=None, b=None, keepdims=False, return_sign=False):
    r"""Compute the log of the sum of exponentials of input elements.

    This function has the same call signature as
    :func:`scipy.special.logsumexp` and mirrors scipy's behaviour except for
    ``-np.inf`` input. If a and b are both ``-inf`` then scipy's function will
    output ``nan`` whereas here we use:

    .. math::

        \lim_{x \to -\infty} x \exp(x) = 0

    Thus, if ``a=-inf`` in ``log(sum(b * exp(a))`` then we can set ``b=0``
    such that that term is ignored in the sum.
    """
    if b is None:
        b = np.ones_like(a)
    b = np.where(a == -np.inf, 0, b)
    return special.logsumexp(a, axis=axis, b=b, keepdims=keepdims,
                             return_sign=return_sign)


def channel_capacity(w):
    r"""Channel capacity (effective sample size).

    .. math::

        H = \sum_i p_i \ln p_i

        p_i = \frac{w_i}{\sum_j w_j}

        N = e^{-H}
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        W = np.array(w)/sum(w)
        H = np.nansum(np.log(W)*W)
        return np.exp(-H)


def compress_weights(w, u=None, ncompress=True):
    """Compresses weights to their approximate channel capacity."""
    if u is None:
        u = np.random.rand(len(w))

    if w is None:
        w = np.ones_like(u)

    if ncompress is True:
        ncompress = channel_capacity(w)
    elif ncompress is False:
        return w

    if ncompress <= 0:
        W = w/w.max()
    else:
        W = w * ncompress / w.sum()

    fraction, integer = np.modf(W)
    extra = (u < fraction).astype(int)
    return (integer + extra).astype(int)


def quantile(a, q, w=None, interpolation='linear'):
    """Compute the weighted quantile for a one dimensional array."""
    if w is None:
        w = np.ones_like(a)
    a = np.array(list(a))  # Necessary to convert pandas arrays
    w = np.array(list(w))  # Necessary to convert pandas arrays
    i = np.argsort(a)
    c = np.cumsum(w[i[1:]]+w[i[:-1]])
    c = c / c[-1]
    c = np.concatenate(([0.], c))
    icdf = interp1d(c, a[i], kind=interpolation)
    quant = icdf(q)
    if isinstance(q, float):
        quant = float(quant)
    return quant


def mirror_1d(d, xmin=None, xmax=None):
    """If necessary apply reflecting boundary conditions."""
    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        return np.concatenate((2*xmin-d[d < xmed], d, 2*xmax-d[d >= xmed]))
    elif xmin is not None:
        return np.concatenate((2*xmin-d, d))
    elif xmax is not None:
        return np.concatenate((d, 2*xmax-d))
    else:
        return d


def mirror_2d(d_x_, d_y_, xmin=None, xmax=None, ymin=None, ymax=None):
    """If necessary apply reflecting boundary conditions."""
    d_x = d_x_.copy()
    d_y = d_y_.copy()

    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        d_y = np.concatenate((d_y[d_x < xmed], d_y, d_y[d_x >= xmed]))
        d_x = np.concatenate((2*xmin-d_x[d_x < xmed], d_x,
                              2*xmax-d_x[d_x >= xmed]))
    elif xmin is not None:
        d_y = np.concatenate((d_y, d_y))
        d_x = np.concatenate((2*xmin-d_x, d_x))
    elif xmax is not None:
        d_y = np.concatenate((d_y, d_y))
        d_x = np.concatenate((d_x, 2*xmax-d_x))

    if ymin is not None and ymax is not None:
        ymed = (ymin+ymax)/2
        d_x = np.concatenate((d_x[d_y < ymed], d_x, d_x[d_y >= ymed]))
        d_y = np.concatenate((2*ymin-d_y[d_y < ymed], d_y,
                              2*ymax-d_y[d_y >= ymed]))
    elif ymin is not None:
        d_x = np.concatenate((d_x, d_x))
        d_y = np.concatenate((2*ymin-d_y, d_y))
    elif ymax is not None:
        d_x = np.concatenate((d_x, d_x))
        d_y = np.concatenate((d_y, 2*ymax-d_y))

    return d_x, d_y


def nest_level(lst):
    """Calculate the nesting level of a list."""
    if not isinstance(lst, list):
        return 0
    if not lst:
        return 1
    return max(nest_level(item) for item in lst) + 1


def histogram(a, **kwargs):
    """Produce a histogram for path-based plotting.

    This is a cheap histogram. Necessary if one wants to update the histogram
    dynamically, and redrawing and filling is very expensive.

    This has the same arguments and keywords as :func:`numpy.histogram`, but is
    normalised to 1.
    """
    hist, bin_edges = np.histogram(a, **kwargs)
    xpath, ypath = np.empty((2, 4*len(hist)))
    ypath[0::4] = ypath[3::4] = 0
    ypath[1::4] = ypath[2::4] = hist
    xpath[0::4] = xpath[1::4] = bin_edges[:-1]
    xpath[2::4] = xpath[3::4] = bin_edges[1:]
    mx = max(ypath)
    if mx:
        ypath /= max(ypath)
    return xpath, ypath


def compute_nlive(death, birth):
    """Compute number of live points from birth and death contours.

    Parameters
    ----------
    death, birth : array-like
        list of birth and death contours

    Returns
    -------
    nlive : np.array
        number of live points at each contour
    """
    b = pandas.DataFrame(np.array(birth), columns=['logL'])
    d = pandas.DataFrame(np.array(death), columns=['logL'],
                         index=b.index + len(b))
    b['n'] = +1
    d['n'] = -1
    t = pandas.concat([b, d]).sort_values(['logL', 'n'])
    n = t.n.cumsum()
    return (n[d.index]+1).to_numpy()


def compute_insertion_indexes(death, birth):
    """Compute the live point insertion index for each point.

    For more detail, see `Fowlie et al. (2020)
    <https://arxiv.org/abs/2006.03371>`_

    Parameters
    ----------
    death, birth : array-like
        list of birth and death contours

    Returns
    -------
    indexes : np.array
        live point index at which each live point was inserted
    """
    indexes = np.zeros_like(birth, dtype=int)
    for i, (b, d) in enumerate(zip(birth, death)):
        i_live = (death > b) & (birth <= b)
        live = death[i_live]
        live.sort()
        indexes[i] = np.searchsorted(live, d)
    return indexes


def unique(a):
    """Find unique elements, retaining order."""
    b = []
    for x in a:
        if x not in b:
            b.append(x)
    return b


def iso_probability_contours(pdf, contours=[0.95, 0.68]):
    """Compute the iso-probability contour values."""
    if len(contours) > 1 and not np.all(contours[:-1] > contours[1:]):
        raise ValueError(
            "The kwargs `levels` and `contours` have to be ordered from "
            "outermost to innermost contour, i.e. in strictly descending "
            "order when referring to the enclosed probability mass, e.g. "
            "like the default [0.95, 0.68]. "
            "This breaking change in behaviour was introduced in version "
            "2.0.0-beta.10, in order to better match the ordering of other "
            "matplotlib kwargs."
        )
    contours = [1-p for p in contours]
    p = np.sort(np.array(pdf).flatten())
    m = np.cumsum(p)
    m /= m[-1]
    interp = interp1d([0]+list(m), [0]+list(p))
    c = list(interp(contours))+[max(p)]

    return c


def iso_probability_contours_from_samples(pdf, contours=[0.95, 0.68],
                                          weights=None):
    """Compute the iso-probability contour values."""
    if len(contours) > 1 and not np.all(contours[:-1] > contours[1:]):
        raise ValueError(
            "The kwargs `levels` and `contours` have to be ordered from "
            "outermost to innermost contour, i.e. in strictly descending "
            "order when referring to the enclosed probability mass, e.g. "
            "like the default [0.95, 0.68]. "
            "This breaking change in behaviour was introduced in version "
            "2.0.0-beta.10, in order to better match the ordering of other "
            "matplotlib kwargs."
        )
    if weights is None:
        weights = np.ones_like(pdf)
    contours = [1-p for p in contours]
    i = np.argsort(pdf)
    m = np.cumsum(weights[i])
    m /= m[-1]
    interp = interp1d([0]+list(m), [0]+list(pdf[i]))
    c = list(interp(contours))+[max(pdf)]

    return c


def scaled_triangulation(x, y, cov):
    """Triangulation scaled by a covariance matrix.

    Parameters
    ----------
    x, y : array-like
        x and y coordinates of samples

    cov : array-like, 2d
        Covariance matrix for scaling

    Returns
    -------
    :class:`matplotlib.tri.Triangulation`
        Triangulation with the appropriate scaling
    """
    L = np.linalg.cholesky(cov)
    Linv = np.linalg.inv(L)
    x_, y_ = Linv.dot([x, y])
    tri = Triangulation(x_, y_)
    return Triangulation(x, y, tri.triangles)


def triangular_sample_compression_2d(x, y, cov, w=None, n=1000):
    """Histogram a 2D set of weighted samples via triangulation.

    This defines bins via a triangulation of the subsamples and sums weights
    within triangles surrounding each point

    Parameters
    ----------
    x, y : array-like
        x and y coordinates of samples for compressing

    cov : array-like, 2d
        Covariance matrix for scaling

    w : :class:`pandas.Series`, optional
        weights of samples

    n : int, default=1000
        number of samples returned.

    Returns
    -------
    tri :
        :class:`matplotlib.tri.Triangulation` with an appropriate scaling

    w : array-like
        Compressed samples and weights
    """
    # Pre-process samples to not be affected by non-standard indexing
    # Details: https://github.com/handley-lab/anesthetic/issues/189
    x = np.array(x)
    y = np.array(y)

    x = pandas.Series(x)
    if w is None:
        w = pandas.Series(index=x.index, data=np.ones_like(x))

    # Select samples for triangulation
    if (w != 0).sum() < n:
        i = x.index
    else:
        i = np.random.choice(x.index, size=n, replace=False, p=w/w.sum())

    # Generate triangulation
    tri = scaled_triangulation(x[i], y[i], cov)

    # For each point find corresponding triangles
    trifinder = tri.get_trifinder()
    j = trifinder(x, y)
    k = tri.triangles[j[j != -1]]

    # Compute mass in each triangle, and add it to each corner
    w_ = np.zeros(len(i))
    for i in range(3):
        np.add.at(w_, k[:, i], w[j != -1]/3)

    return tri, w_


def sample_compression_1d(x, w=None, ncompress=True):
    """Histogram a 1D set of weighted samples via subsampling.

    This compresses the number of samples, combining weights.

    Parameters
    ----------
    x : array-like
        x coordinate of samples for compressing

    w : :class:`pandas.Series`, optional
        weights of samples

    ncompress : int, default=True
        Degree of compression.

        * If int: number of samples returned.
        * If True: compresses to the channel capacity.
        * If False: no compression.

    Returns
    -------
    x, w: array-like
        Compressed samples and weights
    """
    if ncompress is False:
        return x, w
    elif ncompress is True:
        ncompress = channel_capacity(w)
    x = np.array(x)
    if w is None:
        w = np.ones_like(x)
    w = np.array(w)

    # Select inner samples for triangulation
    if len(x) > ncompress:
        x_ = np.random.choice(x, size=ncompress, replace=False)
    else:
        x_ = x.copy()
    x_.sort()

    # Compress mass onto these subsamples
    centers = (x_[1:] + x_[:-1])/2
    j = np.digitize(x, centers)
    w_ = np.zeros_like(x_)
    np.add.at(w_, j, w)

    return x_, w_


def is_int(x):
    """Test whether x is an integer."""
    return isinstance(x, int) or isinstance(x, np.integer)


def match_contour_to_contourf(contours, vmin, vmax):
    """Get needed `vmin, vmax` to match `contour` colors to `contourf` colors.

    `contourf` uses the arithmetic mean of contour levels to assign colors,
    whereas `contour` uses the contour level directly. To get the same colors
    for `contour` lines as for `contourf` faces, we need some fiddly algebra.
    """
    if len(contours) <= 2:
        vmin = 2 * vmin - vmax
        return vmin, vmax
    else:
        c0 = contours[0]
        c1 = contours[1]
        ce = contours[-2]
        denom = vmax + ce - c1 - c0
        vmin = +(c0 * vmax - c1 * ce + 2 * vmin * (ce - c0)) / denom
        vmax = -(c0 * vmax + c1 * ce - 2 * vmax * ce) / denom
        return vmin, vmax


def insertion_p_value(indexes, nlive, batch=0):
    """Compute the p-value from insertion indexes, assuming constant nlive.

    Note that this function doesn't use :func:`scipy.stats.kstest` as the
    latter assumes continuous distributions.

    For more detail, see `Fowlie et al. (2020)
    <https://arxiv.org/abs/2006.03371>`_

    For a rolling test, you should provide the optional parameter ``batch!=0``.
    In this case the test computes the p-value on consecutive batches of size
    ``nlive * batch``, selects the smallest one and adjusts for multiple
    comparisons using a Bonferroni correction.

    Parameters
    ----------
    indexes : array-like
        list of insertion indexes, sorted by death contour

    nlive : int
        number of live points

    batch : float
        batch size in units of nlive for a rolling p-value

    Returns
    -------
    ks_result : dict
        Kolmogorov-Smirnov test results:

            * ``D``: Kolmogorov-Smirnov statistic
            * ``sample_size``: sample size
            * ``p-value``: p-value

            if ``batch != 0``:

            * ``iterations``: bounds of batch with minimum p-value
            * ``nbatches``: the number of batches in total
            * ``uncorrected p-value``: p-value without Bonferroni correction
    """
    if batch == 0:
        bins = np.arange(-0.5, nlive + 0.5, 1.)
        empirical_pmf = np.histogram(
            np.array(indexes),
            bins=bins,
            density=True,
        )[0]
        empirical_cmf = np.cumsum(empirical_pmf)
        uniform_cmf = np.arange(1., nlive + 1., 1.) / nlive

        D = abs(empirical_cmf - uniform_cmf).max()
        sample_size = len(indexes)
        K = D * np.sqrt(sample_size)

        ks_result = {}
        ks_result["D"] = D
        ks_result["sample_size"] = sample_size
        ks_result["p-value"] = kstwobign.sf(K)
        return ks_result
    else:
        batch = int(batch * nlive)
        batches = [
            np.array(indexes)[i:i + batch]
            for i in range(0, len(indexes), batch)
        ]
        ks_results = [insertion_p_value(c, nlive) for c in batches]
        ks_result = min(ks_results, key=lambda t: t["p-value"])
        index = ks_results.index(ks_result)

        ks_result["iterations"] = (index * batch, (index + 1) * batch)
        ks_result["nbatches"] = n = len(batches)
        ks_result["uncorrected p-value"] = p = ks_result["p-value"]
        ks_result["p-value"] = 1. - (1. - p)**n if p != 1 else p * n
        return ks_result


@contextlib.contextmanager
def temporary_seed(seed):
    """Context for temporarily setting a numpy seed."""
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def adjust_docstrings(cls, pattern, repl, *args, **kwargs):
    """Adjust the docstrings of a class using regular expressions.

    After the first argument, the remaining arguments are identical to re.sub.

    Parameters
    ----------
    cls : class
        class to adjust

    pattern : str
        regular expression pattern

    repl : str
        replacement string
    """
    for key, val in cls.__dict__.items():
        doc = inspect.getdoc(val)
        if doc is not None:
            newdoc = re.sub(pattern, repl, doc, *args, **kwargs)
            try:
                cls.__dict__[key].__doc__ = newdoc
            except AttributeError:
                pass



"""Kernel density estimation tools.

These act as a wrapper around fastKDE, but could be replaced in future by
alternative kernel density estimators
"""
import warnings
from fastkde import fastKDE
from anesthetic.utils import mirror_1d, mirror_2d


def fastkde_1d(d, xmin=None, xmax=None):
    """Perform a one-dimensional kernel density estimation.

    Wrapper around `fastkde.fastKDE <https://github.com/LBL-EESA/fastkde>`_.
    Boundary corrections implemented by reflecting boundary conditions.

    Parameters
    ----------
    d : np.array
        Data to perform kde on

    xmin, xmax : float
        lower/upper prior bounds
        optional, default None

    Returns
    -------
    x : np.array
        x-coordinates of kernel density estimates
    p : np.array
        kernel density estimates

    """
    f = xmax is None or xmin is None
    d_ = mirror_1d(d, xmin, xmax)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, x = fastKDE.pdf(d_, axisExpansionFactor=f,
                           numPointsPerSigma=10*(2-f))
    p *= 2-f

    if xmin is not None:
        p = p[x >= xmin]
        x = x[x >= xmin]

    if xmax is not None:
        p = p[x <= xmax]
        x = x[x <= xmax]

    return x, p, xmin, xmax


def fastkde_2d(d_x, d_y, xmin=None, xmax=None, ymin=None, ymax=None):
    """Perform a two-dimensional kernel density estimation.

    Wrapper round `fastkde.fastKDE <https://github.com/LBL-EESA/fastkde>`_.
    Boundary corrections implemented by reflecting boundary conditions.

    Parameters
    ----------
    d_x, d_y : np.array
        x/y coordinates of data to perform kde on

    xmin, xmax, ymin, ymax : float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    x, y : np.array
        x/y-coordinates of kernel density estimates. One-dimensional array
    p : np.array
        kernel density estimates. Two-dimensional array

    """
    f = [xmax is None or xmin is None,
         ymax is None or ymin is None]
    d_x_, d_y_ = mirror_2d(d_x, d_y, xmin, xmax, ymin, ymax)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, (x, y) = fastKDE.pdf(d_x_, d_y_, axisExpansionFactor=f,
                                numPointsPerSigma=10*(2-f[0])*(2-f[1]))

    p *= (2-f[0])
    p *= (2-f[1])
    if xmin is not None:
        p = p[:, x >= xmin]
        x = x[x >= xmin]

    if xmax is not None:
        p = p[:, x <= xmax]
        x = x[x <= xmax]

    if ymin is not None:
        p = p[y >= ymin, :]
        y = y[y >= ymin]

    if ymax is not None:
        p = p[y <= ymax, :]
        y = y[y <= ymax]

    return x, y, p, xmin, xmax, ymin, ymax



"""anesthetic override of matplotlib backend for weighted samples."""
from pandas.plotting._matplotlib import (  # noqa: F401
        PLOT_CLASSES,
        __name__,
        __all__,
        plot,
        register,
        deregister
)
from anesthetic.plotting._matplotlib.boxplot import (  # noqa: F401
    BoxPlot,
    boxplot,
    boxplot_frame,
)
from pandas.plotting._matplotlib import boxplot_frame_groupby  # noqa: F401
from anesthetic.plotting._matplotlib.core import (  # noqa: F401
    AreaPlot,
    BarhPlot,
    BarPlot,
    HexBinPlot,
    LinePlot,
    PiePlot,
    ScatterPlot,
    ScatterPlot2d,
)
from anesthetic.plotting._matplotlib.hist import (  # noqa: F401
    KdePlot,
    Kde1dPlot,
    FastKde1dPlot,
    Kde2dPlot,
    FastKde2dPlot,
    HistPlot,
    Hist1dPlot,
    Hist2dPlot,
    hist_frame,
    hist_series,
)
from anesthetic.plotting._matplotlib.misc import (  # noqa: F401
    bootstrap_plot,
    scatter_matrix,
)
from pandas.plotting._matplotlib import (  # noqa: F401
    andrews_curves,
    autocorrelation_plot,
    lag_plot,
    parallel_coordinates,
    radviz,
    table,
)

PLOT_CLASSES["line"] = LinePlot
PLOT_CLASSES["bar"] = BarPlot
PLOT_CLASSES["barh"] = BarhPlot
PLOT_CLASSES["box"] = BoxPlot
PLOT_CLASSES["hist"] = HistPlot
PLOT_CLASSES["kde"] = KdePlot
PLOT_CLASSES["area"] = AreaPlot
PLOT_CLASSES["pie"] = PiePlot
PLOT_CLASSES["scatter"] = ScatterPlot
PLOT_CLASSES["hexbin"] = HexBinPlot

PLOT_CLASSES['hist_1d'] = Hist1dPlot
PLOT_CLASSES['kde_1d'] = Kde1dPlot
PLOT_CLASSES['fastkde_1d'] = FastKde1dPlot
PLOT_CLASSES['hist_2d'] = Hist2dPlot
PLOT_CLASSES['kde_2d'] = Kde2dPlot
PLOT_CLASSES['fastkde_2d'] = FastKde2dPlot
PLOT_CLASSES['scatter_2d'] = ScatterPlot2d



"""Command-line scripts for anesthetic."""
import argparse
import matplotlib.pyplot as plt
from anesthetic import read_chains


def gui(args=None):
    """Launch the anesthetic GUI.

    See :class:`anesthetic.gui.plot.RunPlotter` for details.
    """
    description = "Nested sampling visualisation"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('file_root', type=str,
                        help="File name root of nested sampling run")
    parser.add_argument('--params', '-p', nargs='*', type=str,
                        help="Parameters to display")
    args = parser.parse_args(args)

    samples = read_chains(root=args.file_root)
    samples.gui(args.params)
    plt.show()



"""Pandas DataFrame and Series with weighted samples."""

import warnings
from inspect import signature
import numpy as np
from pandas import Series, DataFrame, concat, MultiIndex
from pandas.core.groupby import GroupBy, SeriesGroupBy, DataFrameGroupBy, ops
from pandas._libs import lib
from pandas._libs.lib import no_default
from pandas.util._exceptions import find_stack_level
from pandas.util import hash_pandas_object
from numpy.ma import masked_array
from anesthetic.utils import (compress_weights, channel_capacity, quantile,
                              temporary_seed, adjust_docstrings)
from pandas.core.dtypes.missing import notna


class WeightedGroupBy(GroupBy):
    """Weighted version of ``pandas.core.groupby.GroupBy``."""

    grouper: ops.BaseGrouper
    """:meta private:"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _add_weights(self, name, *args, **kwargs):
        result = self.agg(lambda df: getattr(self.obj._constructor(df), name)
                          (*args, **kwargs)).set_weights(self.get_weights())
        return result.__finalize__(self.obj, method="groupby")

    def mean(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("mean", *args, **kwargs)

    def std(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("std", *args, **kwargs)

    def median(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("median", *args, **kwargs)

    def var(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("var", *args, **kwargs)

    def kurt(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("kurt", *args, **kwargs)

    def kurtosis(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("kurtosis", *args, **kwargs)

    def sem(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("sem", *args, **kwargs)

    def skew(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("skew", *args, **kwargs)

    def quantile(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("quantile", *args, **kwargs)

    def get_weights(self):
        """Return the weights of the grouped samples."""
        return self.agg(lambda df: df.get_weights().sum())

    def _op_via_apply(self, name, *args, **kwargs):
        result = super()._op_via_apply(name, *args, **kwargs)
        try:
            index = result.index.get_level_values(self.keys)
            weights = self.get_weights()[index]
        except KeyError:
            weights = self.get_weights()
        return result.set_weights(weights, level=1)


class WeightedSeriesGroupBy(WeightedGroupBy, SeriesGroupBy):
    """Weighted version of ``pandas.core.groupby.SeriesGroupBy``."""

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.obj.get_weights(), *args, **kwargs)

    def cov(self, *args, **kwargs):  # noqa: D102
        return self._op_via_apply("cov", *args, **kwargs)


class WeightedDataFrameGroupBy(WeightedGroupBy, DataFrameGroupBy):
    """Weighted version of ``pandas.core.groupby.DataFrameGroupBy``."""

    def get_weights(self):
        """Return the weights of the grouped samples."""
        return super().get_weights().min(axis=1-self.axis)

    def _gotitem(self, key, ndim: int, subset=None):  # pragma: no cover
        if ndim == 2:
            if subset is None:
                subset = self.obj
            return WeightedDataFrameGroupBy(
                subset,
                self.grouper,
                axis=self.axis,
                level=self.level,
                grouper=self.grouper,
                exclusions=self.exclusions,
                selection=key,
                as_index=self.as_index,
                sort=self.sort,
                group_keys=self.group_keys,
                observed=self.observed,
                mutated=self.mutated,
                dropna=self.dropna,
            )
        elif ndim == 1:
            if subset is None:
                subset = self.obj[key]
            return WeightedSeriesGroupBy(
                subset,
                level=self.level,
                grouper=self.grouper,
                selection=key,
                sort=self.sort,
                group_keys=self.group_keys,
                observed=self.observed,
                dropna=self.dropna,
            )

        raise AssertionError("invalid ndim for _gotitem")

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.obj.get_weights(), *args, **kwargs)

    def cov(self, *args, **kwargs):  # noqa: D102
        return self._op_via_apply("cov", *args, **kwargs)


class _WeightedObject(object):
    """Common methods for `WeightedSeries` and `WeightedDataFrame`.

    :meta public:
    """

    def __init__(self, *args, **kwargs):
        weights = kwargs.pop('weights', None)
        super().__init__(*args, **kwargs)
        if weights is not None:
            self.set_weights(weights, inplace=True)

    def isweighted(self, axis=0):
        """Determine if weights are actually present."""
        return 'weights' in self._get_axis(axis).names

    def get_weights(self, axis=0):
        """Retrieve sample weights from an axis."""
        if self.isweighted(axis):
            return self._get_axis(axis).get_level_values('weights').to_numpy()
        else:
            return np.ones_like(self._get_axis(axis))

    def drop_weights(self, axis=0):
        """Drop weights."""
        if self.isweighted(axis):
            return self.droplevel('weights', axis)
        return self.copy().__finalize__(self, "drop_weights")

    def set_weights(self, weights, axis=0, inplace=False, level=None):
        """Set sample weights along an axis.

        Parameters
        ----------
        weights : 1d array-like
            The sample weights to put in an index.

        axis : int (0,1), default=0
            Whether to put weights in an index or column.

        inplace : bool, default=False
            Whether to operate inplace, or return a new array.

        level : int
            Which level in the index to insert before.
            Defaults to inserting at back

        """
        if inplace:
            result = self
        else:
            result = self.copy()

        if weights is None:
            if result.isweighted(axis=axis):
                result = result.drop_weights(axis)
        else:
            names = [n for n in result._get_axis(axis).names if n != 'weights']
            index = [result._get_axis(axis).get_level_values(n) for n in names]
            if level is None:
                if result.isweighted(axis):
                    level = result._get_axis(axis).names.index('weights')
                else:
                    level = len(index)
            index.insert(level, weights)
            names.insert(level, 'weights')

            index = MultiIndex.from_arrays(index, names=names)
            result = result.set_axis(index, axis=axis, copy=False)

        if inplace:
            self._update_inplace(result)
        else:
            return result.__finalize__(self, "set_weights")

    def _rand(self, axis=0):
        """Random number for consistent compression."""
        seed = hash_pandas_object(self._get_axis(axis)).sum() % 2**32
        with temporary_seed(seed):
            return np.random.rand(self.shape[axis])

    def reset_index(self, level=None, drop=False, inplace=False,
                    *args, **kwargs):
        """Reset the index, retaining weights."""
        weights = self.get_weights()
        answer = super().reset_index(level=level, drop=drop,
                                     inplace=False, *args, **kwargs)
        answer.set_weights(weights, inplace=True)
        if inplace:
            self._update_inplace(answer)
        else:
            return answer.__finalize__(self, "reset_index")

    def neff(self, axis=0):
        """Effective number of samples."""
        if self.isweighted(axis):
            return channel_capacity(self.get_weights(axis))
        else:
            return self.shape[axis]


class WeightedSeries(_WeightedObject, Series):
    """Weighted version of :class:`pandas.Series`."""

    def mean(self, skipna=True):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        null = self.isnull() & skipna
        return np.average(masked_array(self, null), weights=self.get_weights())

    def std(self, *args, **kwargs):  # noqa: D102
        return np.sqrt(self.var(*args, **kwargs))

    def kurtosis(self, *args, **kwargs):  # noqa: D102
        return self.kurt(*args, **kwargs)

    def median(self, *args, **kwargs):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        return self.quantile(*args, **kwargs)

    def var(self, skipna=True):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        null = self.isnull() & skipna
        mean = self.mean(skipna=skipna)
        if np.isnan(mean):
            return mean
        return np.average(masked_array((self-mean)**2, null),
                          weights=self.get_weights())

    def cov(self, other, *args, **kwargs):  # noqa: D102

        this, other = self.align(other, join="inner", copy=False)
        if len(this) == 0:
            return np.nan

        weights = self.index.to_frame()['weights']
        weights, _ = weights.align(other, join="inner", copy=False)

        valid = notna(this) & notna(other)
        if not valid.all():
            this = this[valid]
            other = other[valid]
            weights = weights[valid]

        return np.cov(this, other, aweights=weights)[0, 1]

    def corr(self, other, *args, **kwargs):  # noqa: D102
        norm = self.std(skipna=True)*other.std(skipna=True)
        return self.cov(other, *args, **kwargs)/norm

    def kurt(self, skipna=True):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        null = self.isnull() & skipna
        mean = self.mean(skipna=skipna)
        std = self.std(skipna=skipna)
        if np.isnan(mean) or np.isnan(std):
            return np.nan
        return np.average(masked_array(((self-mean)/std)**4, null),
                          weights=self.get_weights())

    def skew(self, skipna=True):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        null = self.isnull() & skipna
        mean = self.mean(skipna=skipna)
        std = self.std(skipna=skipna)
        if np.isnan(mean) or np.isnan(std):
            return np.nan
        return np.average(masked_array(((self-mean)/std)**3, null),
                          weights=self.get_weights())

    def mad(self, skipna=True):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        null = self.isnull() & skipna
        mean = self.mean(skipna=skipna)
        if np.isnan(mean):
            return np.nan
        return np.average(masked_array(abs(self-mean), null),
                          weights=self.get_weights())

    def sem(self, skipna=True):  # noqa: D102
        return np.sqrt(self.var(skipna=skipna)/self.neff())

    def quantile(self, q=0.5, interpolation='linear'):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        return quantile(self.to_numpy(), q, self.get_weights(), interpolation)

    def compress(self, ncompress=True):
        """Reduce the number of samples by discarding low-weights.

        Parameters
        ----------
        ncompress : int, optional
            effective number of samples after compression. If not supplied
            (or True), then reduce to the channel capacity (theoretical optimum
            compression). If <=0, then compress so that all weights are unity.

        """
        i = compress_weights(self.get_weights(), self._rand(), ncompress)
        return self.repeat(i)

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.get_weights(), *args, **kwargs)

    @property
    def _constructor(self):
        return WeightedSeries

    @property
    def _constructor_expanddim(self):
        return WeightedDataFrame

    def groupby(
        self,
        by=None,
        axis=0,
        level=None,
        as_index=True,
        sort=True,
        group_keys=True,
        observed=False,
        dropna=True,
    ):  # pragma: no cover  # noqa: D102
        if level is None and by is None:
            raise TypeError("You have to supply one of 'by' and 'level'")
        if not as_index:
            raise TypeError("as_index=False only valid with DataFrame")
        axis = self._get_axis_number(axis)

        return WeightedSeriesGroupBy(
            obj=self,
            keys=by,
            axis=axis,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna,
        )


class WeightedDataFrame(_WeightedObject, DataFrame):
    """Weighted version of :class:`pandas.DataFrame`."""

    def mean(self, axis=0, skipna=True, *args, **kwargs):  # noqa: D102
        if self.isweighted(axis):
            if self.get_weights(axis).sum() == 0:
                return self._constructor_sliced(np.nan,
                                                index=self._get_axis(1-axis))
            null = self.isnull() & skipna
            mean = np.average(masked_array(self, null),
                              weights=self.get_weights(axis), axis=axis)
            return self._constructor_sliced(mean, index=self._get_axis(1-axis))
        else:
            return super().mean(axis=axis, skipna=skipna, *args, **kwargs)

    def std(self, *args, **kwargs):  # noqa: D102
        return np.sqrt(self.var(*args, **kwargs))

    def kurtosis(self, *args, **kwargs):  # noqa: D102
        return self.kurt(*args, **kwargs)

    def median(self, *args, **kwargs):  # noqa: D102
        return self.quantile(*args, **kwargs)

    def var(self, axis=0, skipna=True, *args, **kwargs):  # noqa: D102
        if self.isweighted(axis):
            if self.get_weights(axis).sum() == 0:
                return self._constructor_sliced(np.nan,
                                                index=self._get_axis(1-axis))
            null = self.isnull() & skipna
            mean = self.mean(axis=axis, skipna=skipna)
            var = np.average(masked_array((self-mean)**2, null),
                             weights=self.get_weights(axis), axis=axis)
            return self._constructor_sliced(var, index=self._get_axis(1-axis))
        else:
            return super().var(axis=axis, skipna=skipna, *args, **kwargs)

    def cov(self, *args, **kwargs):  # noqa: D102
        if self.isweighted():
            null = self.isnull()
            mean = self.mean(skipna=True)
            x = masked_array(self - mean, null)
            cov = np.ma.dot(self.get_weights()*x.T, x) \
                / self.get_weights().sum().T
            if kwargs:
                raise NotImplementedError("The keywords %s are not implemented"
                                          "for the calculation of the"
                                          "covariance with weighted samples."
                                          % kwargs)
            return self._constructor(cov, index=self.columns,
                                     columns=self.columns)
        else:
            return super().cov(*args, **kwargs)

    def corr(self, method="pearson", skipna=True,
             *args, **kwargs):  # noqa: D102
        if self.isweighted():
            cov = self.cov()
            diag = np.sqrt(np.diag(cov))
            return cov.divide(diag, axis=1).divide(diag, axis=0)
        else:
            return super().corr(*args, **kwargs)

    def corrwith(self, other, axis=0, drop=False, method="pearson",
                 *args, **kwargs):  # noqa: D102
        axis = self._get_axis_number(axis)
        if not self.isweighted(axis):
            return super().corrwith(other, drop=drop, axis=axis, method=method,
                                    *args, **kwargs)
        else:
            if isinstance(other, Series):
                answer = self.apply(lambda x: other.corr(x, method=method),
                                    axis=axis)
                return self._constructor_sliced(answer)

            left, right = self.align(other, join="inner", copy=False)

            if axis == 1:
                left = left.T
                right = right.T

            weights = left.index.to_frame()['weights']
            weights, _ = weights.align(right, join="inner", copy=False)

            # mask missing values
            left = left + right * 0
            right = right + left * 0

            # demeaned data
            ldem = left - left.mean()
            rdem = right - right.mean()

            num = (ldem * rdem * weights.to_numpy()[:, None]).sum()
            dom = weights.sum() * left.std() * right.std()

            correl = num / dom

            if not drop:
                # Find non-matching labels along the given axis
                result_index = self._get_axis(1-axis).union(
                    other._get_axis(1-axis))
                idx_diff = result_index.difference(correl.index)

                if len(idx_diff) > 0:
                    correl = concat([correl, Series([np.nan] * len(idx_diff),
                                                    index=idx_diff)])

            return self._constructor_sliced(correl)

    def kurt(self, axis=0, skipna=True, *args, **kwargs):  # noqa: D102
        if self.isweighted(axis):
            if self.get_weights(axis).sum() == 0:
                return self._constructor_sliced(np.nan,
                                                index=self._get_axis(1-axis))
            null = self.isnull() & skipna
            mean = self.mean(axis=axis, skipna=skipna)
            std = self.std(axis=axis, skipna=skipna)
            kurt = np.average(masked_array(((self-mean)/std)**4, null),
                              weights=self.get_weights(axis), axis=axis)
            return self._constructor_sliced(kurt, index=self._get_axis(1-axis))
        else:
            return super().kurt(axis=axis, skipna=skipna, *args, **kwargs)

    def skew(self, axis=0, skipna=True, *args, **kwargs):  # noqa: D102
        if self.isweighted(axis):
            if self.get_weights(axis).sum() == 0:
                return self._constructor_sliced(np.nan,
                                                index=self._get_axis(1-axis))
            null = self.isnull() & skipna
            mean = self.mean(axis=axis, skipna=skipna)
            std = self.std(axis=axis, skipna=skipna)
            skew = np.average(masked_array(((self-mean)/std)**3, null),
                              weights=self.get_weights(axis), axis=axis)
            return self._constructor_sliced(skew, index=self._get_axis(1-axis))
        else:
            return super().skew(axis=axis, skipna=skipna, *args, **kwargs)

    def mad(self, axis=0, skipna=True, *args, **kwargs):  # noqa: D102
        if self.isweighted(axis):
            if self.get_weights(axis).sum() == 0:
                return self._constructor_sliced(np.nan,
                                                index=self._get_axis(1-axis))
            null = self.isnull() & skipna
            mean = self.mean(axis=axis, skipna=skipna)
            mad = np.average(masked_array(abs(self-mean), null),
                             weights=self.get_weights(axis), axis=axis)
            return self._constructor_sliced(mad, index=self._get_axis(1-axis))
        else:
            return super().var(axis=axis, skipna=skipna, *args, **kwargs)

    def sem(self, axis=0, skipna=True):  # noqa: D102
        n = self.neff(axis)
        return np.sqrt(self.var(axis=axis, skipna=skipna)/n)

    def quantile(self, q=0.5, axis=0, numeric_only=None,
                 interpolation='linear', method=None):  # noqa: D102
        if self.isweighted(axis):
            if numeric_only is not None or method is not None:
                raise NotImplementedError(
                    "`numeric_only` and `method` kwargs not implemented for "
                    "`WeightedSeries` and `WeightedDataFrame`."
                )
            data = np.array([c.quantile(q=q, interpolation=interpolation)
                             for _, c in self.items()])
            if np.isscalar(q):
                return self._constructor_sliced(data,
                                                index=self._get_axis(1-axis))
            else:
                return self._constructor(data.T, index=q,
                                         columns=self._get_axis(1-axis))
        else:
            if numeric_only is None:
                numeric_only = True
            if method is None:
                method = 'single'
            return super().quantile(q=q, axis=axis, numeric_only=numeric_only,
                                    interpolation=interpolation, method=method)

    def compress(self, ncompress=True, axis=0):
        """Reduce the number of samples by discarding low-weights.

        Parameters
        ----------
        ncompress : int, optional
            effective number of samples after compression. If not supplied
            (or True), then reduce to the channel capacity (theoretical optimum
            compression). If <=0, then compress so that all weights are unity.

        """
        if self.isweighted(axis):
            i = compress_weights(self.get_weights(axis), self._rand(axis),
                                 ncompress)
            data = np.repeat(self.to_numpy(), i, axis=axis)
            i = self.drop_weights(axis)._get_axis(axis).repeat(i)
            df = self._constructor(data=data)
            df = df.set_axis(i, axis=axis, copy=False)
            df = df.set_axis(self._get_axis(1-axis), axis=1-axis, copy=False)
            return df
        else:
            return self

    def sample(self, *args, **kwargs):  # noqa: D102
        sig = signature(DataFrame.sample)
        axis = sig.bind(self, *args, **kwargs).arguments.get('axis', 0)
        if self.isweighted(axis):
            return super().sample(weights=self.get_weights(axis),
                                  *args, **kwargs)
        else:
            return super().sample(*args, **kwargs)

    @property
    def _constructor_sliced(self):
        return WeightedSeries

    @property
    def _constructor(self):
        return WeightedDataFrame

    def groupby(
        self,
        by=None,
        axis=no_default,
        level=None,
        as_index: bool = True,
        sort: bool = True,
        group_keys: bool = True,
        observed: bool = False,
        dropna: bool = True,
    ):  # pragma: no cover  # noqa: D102
        if axis is not lib.no_default:
            axis = self._get_axis_number(axis)
            if axis == 1:
                warnings.warn(
                    "DataFrame.groupby with axis=1 is deprecated. Do "
                    "`frame.T.groupby(...)` without axis instead.",
                    FutureWarning,
                    stacklevel=find_stack_level(),
                )
            else:
                warnings.warn(
                    "The 'axis' keyword in DataFrame.groupby is deprecated "
                    "and will be removed in a future version.",
                    FutureWarning,
                    stacklevel=find_stack_level(),
                )
        else:
            axis = 0

        if level is None and by is None:
            raise TypeError("You have to supply one of 'by' and 'level'")

        return WeightedDataFrameGroupBy(
            obj=self,
            keys=by,
            axis=axis,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna,
        )


for cls in [WeightedDataFrame, WeightedSeries, WeightedGroupBy,
            WeightedDataFrameGroupBy, WeightedSeriesGroupBy]:
    adjust_docstrings(cls, r'\bDataFrame\b', 'WeightedDataFrame')
    adjust_docstrings(cls, r'\bDataFrames\b', 'WeightedDataFrames')
    adjust_docstrings(cls, r'\bSeries\b', 'WeightedSeries')
    adjust_docstrings(cls, 'core', 'pandas.core')
    adjust_docstrings(cls, 'pandas.core.window.Rolling.quantile',
                           'pandas.core.window.rolling.Rolling.quantile')
    adjust_docstrings(cls, r'\bDataFrameGroupBy\b', 'WeightedDataFrameGroupBy')
    adjust_docstrings(cls, r'\bSeriesGroupBy\b', 'WeightedSeriesGroupBy')
adjust_docstrings(WeightedDataFrame, 'resample', 'pandas.DataFrame.resample')
adjust_docstrings(WeightedSeries,    'resample', 'pandas.Series.resample')



"""Utility functions for nested sampling examples."""
import numpy as np
from scipy.stats import special_ortho_group
from scipy.special import gamma, gammaln


def random_ellipsoid(mean, cov, size=None):
    """Draw a point uniformly in an ellipsoid.

    This is defined so that the volume of the ellipsoid is
    ``sqrt(det(cov))*V_n``. where ``V_n`` is the volume of the unit n ball,
    and so that its axes have length equal to the square root of the
    eigenvalues of the covariance matrix.

    Parameters
    ----------
    mean : 1d array-like
        The center of mass of the ellipsoid

    cov : 2d array-like
        The covariance structure of the ellipsoid. Axes have lengths equal to
        the square root of the eigenvalues of this matrix.

    size : int or tuple of ints, optional
        Given a shape of, for example, (m,n,k), m*n*k samples are generated,
        and packed in an m-by-n-by-k arrangement. Because each sample is
        N-dimensional, the output shape is (m,n,k,N). If no shape is specified,
        a single (N-D) sample is returned.

    Returns
    -------
    points : array-like
        The drawn samples, of shape size, if that was provided. If not, the
        shape is (N,).
        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional value
        drawn uniformly from the ellipsoid.
    """
    d = len(mean)
    L = np.linalg.cholesky(cov)
    x = np.random.multivariate_normal(np.zeros(d), np.eye(d), size=size)
    r = np.linalg.norm(x, axis=-1, keepdims=True)
    u = np.random.power(d, r.shape)
    return mean + (u*x/r) @ L.T


def random_covariance(sigmas):
    """Draw a randomly oriented covariance matrix with axes length sigmas.

    Parameters
    ----------
    sigmas : 1d array like
        Lengths of the axes of the ellipsoid.

    Returns
    -------
    Covariance matrix : 2d np.array
        shape (len(sigmas), len(sigmas)).
    """
    d = len(sigmas)
    R = special_ortho_group.rvs(d)
    return R @ np.diag(sigmas) @ np.diag(sigmas) @ R.T


def volume_n_ball(n, r=1):
    """Volume of an n dimensional ball, radius r."""
    return np.pi**(n/2)/gamma(1+n/2)*r**n


def log_volume_n_ball(n, r=1):
    """Log-volume of an n dimensional ball, radius r."""
    return np.log(np.pi)*n/2 - gammaln(1+n/2) + np.log(r)*n



import numpy as np
from numpy.testing import assert_array_equal
from anesthetic.labelled_pandas import LabelledSeries, LabelledDataFrame
from pandas import Series, DataFrame, MultiIndex
import pandas.testing
import pytest


def assert_series_equal(x, y):
    pandas.testing.assert_series_equal(Series(x), Series(y))


def assert_series_equal_not_index(x, y):
    assert_series_equal(x.drop_labels(), y)
    with pytest.raises(AssertionError):
        assert_series_equal(x, y)


def assert_series_equal_not_name(x, y):
    with pytest.raises(AssertionError):
        assert_series_equal(x, y)
    x.name = y.name
    assert_series_equal(x, y)


def assert_frame_equal(x, y):
    pandas.testing.assert_frame_equal(DataFrame(x), DataFrame(y))


def assert_frame_equal_not_index(x, y):
    assert_frame_equal(x.drop_labels([0, 1]), y)
    with pytest.raises(AssertionError):
        assert_frame_equal(x, y)


@pytest.fixture
def lseries():
    index = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    labels = ['$%s$' % i for i in index]
    data = np.random.rand(len(index))

    lseries = LabelledSeries(data, index, labels=labels)
    series = Series(data, index)

    assert_series_equal_not_index(lseries, series)
    isinstance(lseries, LabelledSeries)

    assert lseries.A == series.A
    assert lseries['A'] == series['A']
    assert lseries.loc['A'] == series.loc['A']
    assert lseries.at['A'] == series.at['A']
    assert lseries.xs('A') == series.xs('A')
    assert_series_equal_not_index(lseries['A':'D'], series['A':'D'])

    assert lseries[('A', '$A$')] == lseries['A']
    assert lseries.loc['A', '$A$'] == lseries['A']
    assert lseries.at['A', '$A$'] == lseries['A']
    assert lseries.xs(('A', '$A$')) == lseries['A']

    for c in index:
        assert lseries.get_labels_map()[c] == '$%s$' % c
        assert lseries.get_label(c) == '$%s$' % c

    return lseries


@pytest.mark.filterwarnings("ignore::pandas.errors.PerformanceWarning")
def test_LabelledSeries_MultiIndex():
    index = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    vowels = ['vowel' if c in 'AEIOU' else 'consonant' for c in index]
    labels = ['$%s$' % i for i in index]
    data = np.random.rand(len(index))

    multiindex = MultiIndex.from_arrays([vowels, index],
                                        names=['vowels', None])
    lseries = LabelledSeries(data, multiindex, labels=labels)
    series = Series(data, multiindex)

    assert_series_equal_not_index(lseries, series)

    assert_series_equal_not_index(lseries.vowel, series.vowel)
    assert lseries.vowel.A == series.vowel.A

    assert lseries['vowel', 'A'] == series['vowel', 'A']
    assert_series_equal_not_index(lseries['vowel'], series['vowel'])

    assert_series_equal_not_index(lseries[:, 'A'], series[:, 'A'])
    assert_series_equal_not_index(lseries['vowel', :], series['vowel', :])

    assert lseries.loc['vowel', 'A'] == series.loc['vowel', 'A']
    assert_series_equal_not_index(lseries.loc['vowel'], series.loc['vowel'])

    assert_series_equal_not_index(lseries.loc[:, 'A'], series.loc[:, 'A'])
    assert_series_equal_not_index(lseries.loc['vowel', :],
                                  series.loc['vowel', :])
    assert lseries.loc['vowel', 'A'] == lseries.loc['vowel', 'A']

    assert lseries.at['vowel', 'A'] == series.at['vowel', 'A']
    assert_series_equal_not_index(lseries.at['vowel'], series.at['vowel'])

    assert (lseries.xs(('vowel', 'A')) == series.xs(('vowel', 'A')))

    assert_series_equal_not_index(lseries.xs('A', level=1),
                                  series.xs('A', level=1))
    assert_series_equal_not_index(lseries.xs('vowel', level=0),
                                  series.xs('vowel', level=0))
    assert_array_equal(lseries.xs(('vowel', 'A'), level=[0, 1]),
                       series.xs(('vowel', 'A'), level=[0, 1]))

    with pytest.raises(KeyError):
        lseries['foo']

    for c, v in zip(index, vowels):
        assert lseries.get_labels_map()[v, c] == '$%s$' % c
        assert lseries.get_label((v, c)) == '$%s$' % c


@pytest.fixture
def lframe_index():
    index = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    labels = ['$%s$' % i for i in index]
    data = np.random.rand(len(index), 4)

    lframe = LabelledDataFrame(data, index, labels=labels)
    frame = DataFrame(data, index)

    assert_frame_equal_not_index(lframe, frame)

    assert_series_equal(lframe.loc['A'], frame.loc['A'])
    assert lframe.at['A', 0] == frame.at['A', 0]
    assert_series_equal(lframe.xs('A'), frame.xs('A'))

    assert_array_equal(lframe.loc[('A', '$A$')], frame.loc['A'])
    assert lframe.at[('A', '$A$'), 0] == frame.at['A', 0]
    assert_array_equal(lframe.xs(('A', '$A$')), frame.xs('A'))

    assert_series_equal(lframe.T['A'], lframe.loc['A'])
    assert_series_equal(lframe.T.A, lframe.loc['A'])
    assert_series_equal(lframe.T.loc[:, 'A'], lframe.loc['A'])
    assert_series_equal(lframe.T.loc[0, 'A'], lframe.loc['A', 0])

    assert_series_equal(lframe.T.at[0, 'A'], lframe.at['A', 0])

    assert_frame_equal_not_index(lframe['A':'C'], frame['A':'C'])
    assert_frame_equal_not_index(lframe.loc['A':'C', 0:3],
                                 frame.loc['A':'C', 0:3])
    assert_frame_equal_not_index(lframe.loc['A':'C', 0:3],
                                 frame.loc['A':'C', 0:3])

    with pytest.raises(KeyError):
        lframe['foo']

    for c in index:
        assert lframe.get_labels_map()[c] == '$%s$' % c
        assert lframe.get_label(c) == '$%s$' % c

    return lframe


@pytest.mark.filterwarnings("ignore::pandas.errors.PerformanceWarning")
def test_LabelledDataFrame_index_MultiIndex():
    index = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    vowels = ['vowel' if c in 'AEIOU' else 'consonant' for c in index]
    labels = ['$%s$' % i for i in index]
    data = np.random.rand(len(index), 4)

    multiindex = MultiIndex.from_arrays([vowels, index],
                                        names=['vowels', None])
    lframe = LabelledDataFrame(data, multiindex, labels=labels)
    frame = DataFrame(data, multiindex)

    assert_frame_equal_not_index(lframe, frame)

    assert_frame_equal_not_index(lframe.loc['vowel'], frame.loc['vowel'])
    assert_series_equal(lframe.loc['vowel', 'A'], frame.loc['vowel', 'A'])
    assert_series_equal_not_name(lframe.loc['vowel', 'A', '$A$'],
                                 frame.loc['vowel', 'A'])

    assert_frame_equal_not_index(lframe.xs('vowel'), frame.xs('vowel'))
    assert_series_equal(lframe.xs(('vowel', 'A')), frame.xs(('vowel', 'A')))
    assert_series_equal_not_name(lframe.xs(('vowel', 'A', '$A$')),
                                 frame.xs(('vowel', 'A')))

    assert_series_equal(lframe.loc[('vowel', 'A'), 0],
                        frame.loc[('vowel', 'A'), 0])
    assert_series_equal(lframe.loc[('vowel', 'A', '$A$'), 0],
                        frame.loc[('vowel', 'A'), 0])
    assert lframe.at[('vowel', 'A'), 0] == frame.at[('vowel', 'A'), 0]
    assert lframe.at[('vowel', 'A', '$A$'), 0] == frame.at[('vowel', 'A'), 0]

    assert_frame_equal(lframe.T['vowel'], lframe.loc['vowel'].T)
    assert_series_equal(lframe.T['vowel', 'A'], lframe.loc['vowel', 'A'])
    assert_series_equal_not_name(lframe.T['vowel', 'A', '$A$'],
                                 lframe.loc['vowel', 'A'])
    assert_frame_equal(lframe.T.vowel, lframe.loc['vowel'].T)
    assert_series_equal_not_name(lframe.T.vowel.A, lframe.loc['vowel', 'A'])
    assert_frame_equal(lframe.T.loc[:, 'vowel'], lframe.loc['vowel'].T)
    assert_series_equal(lframe.T.loc[:, ('vowel', 'A')],
                        lframe.loc[('vowel', 'A')])
    assert_series_equal_not_name(lframe.T.loc[:, ('vowel', 'A', '$A$')],
                                 lframe.loc[('vowel', 'A')])
    assert_series_equal(lframe.T.loc[0, 'vowel'], lframe.loc['vowel', 0])
    assert lframe.T.loc[0, ('vowel', 'A')] == lframe.loc[('vowel', 'A'), 0]
    assert (lframe.T.loc[0, ('vowel', 'A', '$A$')]
            == lframe.loc[('vowel', 'A'), 0])

    assert_series_equal(lframe.T.at[0, ('vowel', 'A')],
                        lframe.at[('vowel', 'A'), 0])
    assert_series_equal(lframe.T.at[0, ('vowel', 'A', '$A$')],
                        lframe.at[('vowel', 'A'), 0])

    with pytest.raises(KeyError):
        lframe['foo']

    for c, v in zip(index, vowels):
        assert lframe.get_labels_map()[v, c] == '$%s$' % c
        assert lframe.get_label((v, c)) == '$%s$' % c


def test_LabelledDataFrame_column():
    columns = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    labels = ['$%s$' % i for i in columns]
    data = np.random.rand(4, len(columns))

    lframe = LabelledDataFrame(data, columns=columns)
    lframe.set_labels(labels, axis=1, inplace=True)
    frame = DataFrame(data, columns=columns)

    assert_frame_equal_not_index(lframe, frame)

    assert_series_equal(lframe.A, frame.A)
    assert_series_equal(lframe['A'], frame['A'])
    assert_series_equal_not_name(lframe['A', '$A$'], frame['A'])

    assert_series_equal(lframe.loc[:, 'A'], frame.loc[:, 'A'])
    assert lframe.at[0, 'A'] == frame.at[0, 'A']
    assert_series_equal(lframe.xs('A', axis=1), frame.xs('A', axis=1))

    assert_array_equal(lframe.loc[:, ('A', '$A$')], frame.loc[:, 'A'])
    assert lframe.at[0, ('A', '$A$')] == frame.at[0, 'A']
    assert_array_equal(lframe.xs(('A', '$A$'), axis=1), frame.xs('A', axis=1))

    assert_series_equal(lframe.T.loc['A'], lframe['A'])
    assert_series_equal(lframe.T.loc['A', 0], lframe.loc[0, 'A'])

    assert_series_equal(lframe.T.at['A', 0], lframe.at[0, 'A'])

    assert_frame_equal_not_index(lframe.loc[0:3, 'A':'C'],
                                 frame.loc[0:3, 'A':'C'])
    assert_frame_equal_not_index(lframe.loc[0:3, 'A':'C'],
                                 frame.loc[0:3, 'A':'C'])

    with pytest.raises(KeyError):
        lframe['foo']

    for c in columns:
        assert lframe.get_labels_map(1)[c] == '$%s$' % c
        assert lframe.get_label(c, 1) == '$%s$' % c


@pytest.mark.filterwarnings("ignore::pandas.errors.PerformanceWarning")
def test_LabelledDataFrame_column_MultiIndex():
    columns = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']
    vowels = ['vowel' if c in 'AEIOU' else 'consonant' for c in columns]
    labels = ['$%s$' % i for i in columns]
    data = np.random.rand(4, len(columns))

    multiindex = MultiIndex.from_arrays([vowels, columns],
                                        names=['vowels', None])
    lframe = LabelledDataFrame(data, columns=multiindex)
    lframe.set_labels(labels, axis=1, inplace=True)
    frame = DataFrame(data, columns=multiindex)

    assert_frame_equal_not_index(lframe, frame)

    assert_frame_equal_not_index(lframe.vowel, frame.vowel)
    assert_series_equal(lframe.vowel.A, frame.vowel.A)

    assert_frame_equal_not_index(lframe['vowel'], frame['vowel'])
    assert_series_equal(lframe['vowel', 'A'], frame['vowel', 'A'])
    assert_series_equal_not_name(lframe['vowel', 'A', '$A$'],
                                 frame['vowel', 'A'])

    assert_frame_equal_not_index(lframe.xs('vowel', axis=1),
                                 frame.xs('vowel', axis=1))
    assert_series_equal(lframe.xs(('vowel', 'A'), axis=1),
                        frame.xs(('vowel', 'A'), axis=1))
    assert_series_equal_not_name(lframe.xs(('vowel', 'A', '$A$'), axis=1),
                                 frame.xs(('vowel', 'A'), axis=1))

    assert_series_equal(lframe.loc[0, ('vowel', 'A')],
                        frame.loc[0, ('vowel', 'A')])
    assert_series_equal(lframe.loc[0, ('vowel', 'A', '$A$')],
                        frame.loc[0, ('vowel', 'A')])
    assert lframe.at[0, ('vowel', 'A')] == frame.at[0, ('vowel', 'A')]
    assert lframe.at[0, ('vowel', 'A', '$A$')] == frame.at[0, ('vowel', 'A')]

    assert_frame_equal(lframe.T.loc['vowel'], lframe['vowel'].T)
    assert_series_equal(lframe.T.loc['vowel', 'A'], lframe['vowel', 'A'])
    assert_series_equal_not_name(lframe.T.loc['vowel', 'A', '$A$'],
                                 lframe['vowel', 'A'])
    assert_series_equal_not_name(lframe.T.loc[('vowel', 'A', '$A$')],
                                 lframe[('vowel', 'A')])
    assert_series_equal(lframe.T.loc['vowel', 0], lframe.loc[0, 'vowel'])
    assert lframe.T.loc[('vowel', 'A'), 0] == lframe.loc[0, ('vowel', 'A')]
    assert (lframe.T.loc[('vowel', 'A', '$A$'), 0]
            == lframe.loc[0, ('vowel', 'A')])

    assert_series_equal(lframe.T.at[('vowel', 'A'), 0],
                        lframe.at[0, ('vowel', 'A')])
    assert_series_equal(lframe.T.at[('vowel', 'A', '$A$'), 0],
                        lframe.at[0, ('vowel', 'A')])

    with pytest.raises(KeyError):
        lframe['foo']

    for c, v in zip(columns, vowels):
        assert lframe.get_labels_map(1)[v, c] == '$%s$' % c
        assert lframe.get_label((v, c), 1) == '$%s$' % c


def test_set_labels(lseries):
    labels = lseries.get_labels()
    labels[1] = '$b$'
    assert_array_equal(lseries.set_labels(labels).get_labels(), labels)
    assert lseries.get_labels()[1] != labels[1]
    lseries.set_labels(labels, inplace=True)
    assert lseries.get_labels()[1] == labels[1]

    assert lseries.drop_labels().get_labels() is None
    assert lseries.get_labels() is not None
    assert lseries.set_labels(None).get_labels() is None
    assert lseries.get_labels() is not None
    lseries.set_labels(None, inplace=True)
    assert lseries.get_labels() is None


def test_constructors(lseries, lframe_index):
    lframe = lframe_index
    assert isinstance(lseries, LabelledSeries)
    assert isinstance(lseries.to_frame(), LabelledDataFrame)
    assert isinstance(lframe, LabelledDataFrame)
    assert isinstance(lframe[0], LabelledSeries)
    assert isinstance(lframe.loc['A'], LabelledSeries)


def test_transpose(lframe_index):
    lframe = lframe_index
    lframe._labels = ("labels0", "labels")
    assert lframe.T._labels == ("labels", "labels0")
    assert lframe.transpose()._labels == ("labels", "labels0")


@pytest.fixture
def test_multiaxis(lframe_index):
    lframe = lframe_index.iloc[:4]
    lframe._labels = ('labels', 'aliases')
    columns = MultiIndex.from_arrays([['one', 'two', 'three', 'four'],
                                      [1, 2, 3, 4]],
                                     names=[None, 'aliases'])
    lframe.columns = columns
    result = lframe.copy()

    assert_array_equal(lframe.get_labels(0), ['$A$', '$B$', '$C$', '$D$'])
    assert_array_equal(lframe.get_labels(1), [1, 2, 3, 4])
    assert_array_equal(lframe.T.get_labels(1), ['$A$', '$B$', '$C$', '$D$'])
    assert_array_equal(lframe.T.get_labels(0), [1, 2, 3, 4])

    assert lframe.islabelled(0)
    assert lframe.islabelled(1)

    assert isinstance(lframe.loc['A'], LabelledSeries)
    assert isinstance(lframe['one'], LabelledSeries)
    lframe._labels = ('labels', 'foo')
    assert isinstance(lframe.loc['A'], LabelledSeries)
    assert isinstance(lframe['one'], LabelledDataFrame)

    assert lframe.islabelled(0)
    assert not lframe.islabelled(1)

    lframe._labels = ('labels', 'foo')
    assert isinstance(lframe['one'], LabelledDataFrame)

    assert lframe.islabelled(0)
    assert not lframe.islabelled(1)

    return result


def test_set_label(lseries):
    assert isinstance(lseries.get_labels_map(), Series)

    nolabels_map = lseries.drop_labels().get_labels_map()
    assert isinstance(nolabels_map, Series)
    assert_array_equal(nolabels_map, '')
    assert_array_equal(nolabels_map.index, lseries.drop_labels().index)

    assert lseries.get_label('A') == '$A$'
    assert lseries.set_label('A', '$a$').get_label('A') == '$a$'
    lseries.set_label('A', '$a$')
    assert lseries.get_label('A') == '$A$'
    lseries.set_label('A', '$a$', inplace=True)
    assert lseries.get_label('A') == '$a$'

    lseries = lseries.drop_labels()
    assert lseries.get_label('A') == ''
    assert lseries.index.nlevels == 1

    lseries.set_label('A', '$A$', inplace=True)
    assert lseries.get_label('A') == '$A$'
    assert lseries.get_label('B') == ''
    assert lseries.index.nlevels == 2

    lseries['Z'] = 1.0
    assert lseries.get_label('Z') == ''


def test_multiaxis_slice(test_multiaxis):
    lframe = test_multiaxis
    assert_series_equal_not_name(lframe['one'], lframe[('one', 1)])
    assert_series_equal_not_name(lframe.loc['A'], lframe.loc[('A', '$A$')])
    assert_series_equal_not_name(lframe.loc[:, 'one'],
                                 lframe.loc[:, ('one', 1)])

    assert_frame_equal(lframe.loc[['A', 'B'], ['one', 'two']],
                       lframe.loc[[('A', '$A$'), ('B', '$B$')],
                                  [('one', 1), ('two', 2)]])

    assert_series_equal_not_name(lframe.loc['A', ['one', 'two']],
                                 lframe.loc[('A', '$A$'),
                                            [('one',  1), ('two', 2)]])
    assert_series_equal_not_name(lframe.loc[['A', 'B'], 'one'],
                                 lframe.loc[[('A', '$A$'), ('B', '$B$')],
                                            ('one',  1)])


def test_reset_index(lframe_index):
    ldf = lframe_index
    assert ldf.reset_index().index.names == [None, 'labels']

    assert not np.array_equal(ldf.reset_index().columns, ldf.columns)
    assert_array_equal(ldf.reset_index(drop=True).columns, ldf.columns)

    new = ldf.copy()
    assert not np.array_equal(new.index, ldf.reset_index())
    new.reset_index(inplace=True)
    assert_array_equal(new.index, ldf.reset_index().index)


def test_drop_labels(lframe_index):
    ldf = lframe_index
    assert ldf.islabelled()
    nolabels = ldf.drop_labels()
    assert not nolabels.islabelled()
    assert_frame_equal_not_index(ldf, nolabels)
    assert_frame_equal(nolabels.drop_labels(), nolabels)
    assert nolabels.drop_labels() is not nolabels
